from enum import Enum, auto
import time
from datetime import datetime, timezone
import jwt
import logging
from oauthlib.oauth2.rfc6749.errors import InsecureTransportError, TokenExpiredError, MissingTokenError
from oauthlib.oauth2.rfc6749.utils import is_secure_transport

import requests

from oauthlib.common import UNICODE_ASCII_CHARACTER_SET, generate_nonce, generate_token
from oauthlib.oauth2.rfc6749.parameters import parse_authorization_code_response, parse_token_response, prepare_grant_uri

from requests.adapters import HTTPAdapter

from weconnect.auth.auth_util import addBearerAuthHeader
from weconnect.errors import AuthentificationError, RetrievalError

from weconnect.elements.helpers.blacklist_retry import BlacklistRetry


LOG = logging.getLogger("weconnect")


class AccessType(Enum):
    NONE = auto()
    ACCESS = auto()
    ID = auto()
    REFRESH = auto()


class OpenIDSession(requests.Session):
    def __init__(self, client_id=None, redirect_uri=None, refresh_url=None, scope=None, token=None, metadata={}, state=None, timeout=None,
                 forceReloginAfter=None, **kwargs):
        super(OpenIDSession, self).__init__(**kwargs)
        self.client_id = client_id
        self.redirect_uri = redirect_uri
        self.refresh_url = refresh_url
        self.scope = scope
        self.state = state or generate_token(length=30, chars=UNICODE_ASCII_CHARACTER_SET)

        self.timeout = timeout
        self._token = None
        self.token = token
        self.metadata = metadata
        self.lastLogin = None
        self.forceReloginAfter = forceReloginAfter

        self._retries = False

    @property
    def forceReloginAfter(self):
        return self._forceReloginAfter

    @forceReloginAfter.setter
    def forceReloginAfter(self, newValue):
        self._forceReloginAfter = newValue
        if newValue is not None and self.lastLogin is None:
            self.lastLogin = time.time()

    @property
    def retries(self):
        return self._retries

    @retries.setter
    def retries(self, newValue):
        self._retries = newValue
        if newValue:
            # Retry on internal server error (500)
            retries = BlacklistRetry(total=newValue,
                                     backoff_factor=0.1,
                                     status_forcelist=[500],
                                     status_blacklist=[429],
                                     raise_on_status=False)
            self.mount('https://', HTTPAdapter(max_retries=retries))

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, newToken):
        if newToken is not None:
            # If new token e.g. after refresh is missing expires_in we assume it is the same than before
            if 'expires_in' not in newToken:
                if self._token is not None and 'expires_in' in self._token:
                    newToken['expires_in'] = self._token['expires_in']
                else:
                    if 'id_token' in newToken:
                        try:
                            meta_data = jwt.decode(newToken['id_token'], options={"verify_signature": False})
                            if 'exp' in meta_data:
                                newToken['expires_at'] = meta_data['exp']
                                expires_at = datetime.fromtimestamp(meta_data['exp'], tz=timezone.utc)
                                newToken['expires_in'] = (expires_at - datetime.now(tz=timezone.utc)).total_seconds()
                            else:
                                newToken['expires_in'] = 3600
                        except jwt.exceptions.DecodeError:
                            # Invalid JWT token, use default expiration
                            LOG.debug("Invalid JWT token provided, using default expiration")
                            newToken['expires_in'] = 3600
                    else:
                        newToken['expires_in'] = 3600
            # It expires_in is set and expires_at is not set we calculate expires_at from expires_in using the current time
            if 'expires_in' in newToken and 'expires_at' not in newToken:
                newToken['expires_at'] = time.time() + int(newToken.get('expires_in'))
        self._token = newToken

    @property
    def accessToken(self):
        if self._token is not None and 'access_token' in self._token:
            return self._token.get('access_token')
        return None

    @accessToken.setter
    def accessToken(self, newValue):
        if self._token is None:
            self._token = {}
        self._token['access_token'] = newValue

    @property
    def refreshToken(self):
        if self._token is not None and 'refresh_token' in self._token:
            return self._token.get('refresh_token')
        return None

    @property
    def idToken(self):
        if self._token is not None and 'id_token' in self._token:
            return self._token.get('id_token')
        return None

    @property
    def tokenType(self):
        if self._token is not None and 'token_type' in self._token:
            return self._token.get('token_type')
        return None

    @property
    def expiresIn(self):
        if self._token is not None and 'expires_in' in self._token:
            return self._token.get('expires_in')
        return None

    @property
    def expiresAt(self):
        if self._token is not None and 'expires_at' in self._token:
            return self._token.get('expires_at')
        return None

    @property
    def authorized(self):
        return bool(self.accessToken)

    @property
    def expired(self):
        return self.expiresAt is not None and self.expiresAt < time.time()

    @property
    def userId(self):
        if 'userId' in self.metadata:
            return self.metadata['userId']
        return None

    @userId.setter
    def userId(self, newUserId):
        self.metadata['userId'] = newUserId

    def login(self):
        self.lastLogin = time.time()

    def refresh(self):
        pass

    def authorizationUrl(self, url, state=None, **kwargs):
        state = state or self.state
        authUrl = prepare_grant_uri(uri=url, client_id=self.client_id, redirect_uri=self.redirect_uri, response_type='code id_token token', scope=self.scope,
                                    state=state, nonce=generate_nonce(), **kwargs)
        return authUrl

    def parseFromFragment(self, authorization_response, state=None):
        state = state or self.state
        self.token = parse_authorization_code_response(authorization_response, state=state)
        return self.token

    def parseFromBody(self, token_response, state=None):
        self.token = parse_token_response(token_response, scope=self.scope)
        return self.token

    def request(  # noqa: C901
        self,
        method,
        url,
        data=None,
        headers=None,
        withhold_token=False,
        access_type=AccessType.ACCESS,
        token=None,
        timeout=None,
        **kwargs
    ):
        """Intercept all requests and add the OAuth 2 token if present."""
        if not is_secure_transport(url):
            raise InsecureTransportError()
        if access_type != AccessType.NONE and not withhold_token:
            if self.forceReloginAfter is not None and self.lastLogin is not None and (self.lastLogin + self.forceReloginAfter) < time.time():
                LOG.debug("Forced new login after %ds", self.forceReloginAfter)
                self.login()
            try:
                url, headers, data = self.addToken(url, body=data, headers=headers, access_type=access_type, token=token)
            # Attempt to retrieve and save new access token if expired
            except TokenExpiredError:
                LOG.info('Token expired')
                self.accessToken = None
                try:
                    self.refresh()
                except AuthentificationError as authError:
                    # Check if this is a "Server requests new authorization" error
                    if 'Server requests new authorization' in str(authError):
                        LOG.warning('Server requests new authorization - clearing tokens and forcing re-login')
                        # Clear all tokens to force fresh login
                        if hasattr(self, 'clear_tokens'):
                            self.clearTokens()
                        else:
                            # Fallback for base class
                            self.token = None
                            self.access_token = None
                            self.refresh_token = None
                            self.id_token = None
                        LOG.info('Authentication failed during refresh - attempting new login')
                        self.login()
                except TokenExpiredError:
                    self.login()
                except MissingTokenError:
                    self.login()
                except RetrievalError:
                    LOG.error('Retrieval Error while refreshing token. Probably the token was invalidated. Trying to do a new login instead.')
                    self.login()
                url, headers, data = self.addToken(url, body=data, headers=headers, access_type=access_type, token=token)
            except MissingTokenError:
                LOG.error('Missing token')
                self.login()
                url, headers, data = self.addToken(url, body=data, headers=headers, access_type=access_type, token=token)

        if timeout is None:
            timeout = self.timeout

        return super(OpenIDSession, self).request(
            method, url, headers=headers, data=data, **kwargs
        )

    def addToken(self, uri, body=None, headers=None, access_type=AccessType.ACCESS, token=None, **kwargs):
        if not is_secure_transport(uri):
            raise InsecureTransportError()

        if token is None:
            if access_type == AccessType.ID:
                if not (self.idToken):
                    raise MissingTokenError(description="Missing id token.")
                token = self.idToken
            elif access_type == AccessType.REFRESH:
                if not (self.refreshToken):
                    raise MissingTokenError(description="Missing refresh token.")
                token = self.refreshToken
            else:
                if not self.authorized:
                    self.login()
                if not (self.accessToken):
                    raise MissingTokenError(description="Missing access token.")
                if self.expired:
                    raise TokenExpiredError()
                token = self.accessToken

        headers = addBearerAuthHeader(token, headers)

        return (uri, headers, body)



import re
import logging
import requests

from weconnect.addressable import AddressableObject, ChangeableAttribute
from weconnect.errors import ControlError, SetterError

LOG = logging.getLogger("weconnect")


class GeneralControls(AddressableObject):
    def __init__(
        self,
        localAddress,
        parent,
    ):
        super().__init__(localAddress=localAddress, parent=parent)
        if self.parent.spin is not None:
            self.spinControl = ChangeableAttribute(localAddress='spin', parent=self, value='', valueType=str, valueSetter=self.__setSPINControlChange,
                                                   valueGetter=self.__getSPINControlChange)
        else:
            self.spinControl = None

    def __setSPINControlChange(self, value):  # noqa: C901
        if value is None:
            self.parent.spin = value
            return
        elif value == 'None' or value == '':
            self.parent.spin = None
            return
        elif not re.match(r"^\d{4}$", value):
            raise ControlError(f'S-PIN {value} cannot be set/verified, needs to be 4 digits')

        url = 'https://emea.bff.cariad.digital/vehicle/v1/spin/verify'

        data = {}
        data['spin'] = value
        controlResponse = self.parent.session.post(url, json=data, allow_redirects=True)
        if controlResponse.status_code != requests.codes['no_content']:
            dataDict = controlResponse.json()
            if dataDict is not None and 'data' in dataDict:
                errorDict = dataDict['data']
                print(errorDict)
                if errorDict is not None and 'error' in errorDict:
                    errormessage = f'Error: {errorDict["error"]["errorType"]}, S-PIN State: {errorDict["error"]["spinState"]}'
                    if errorDict["error"]["remainingTries"] > 0:
                        errormessage += f', Remaining tries: {errorDict["error"]["remainingTries"]}'
                    if errorDict["error"]["spinLockedWaitingTime"] > 0:
                        errormessage += f', Locked waiting time: {errorDict["error"]["spinLockedWaitingTime"]}'
                    raise SetterError(errormessage)
        self.parent.spin = value

    def __getSPINControlChange(self):
        return 'For security reasons the S-PIN cannot be read out'



import json
import re
from html.parser import HTMLParser


def addBearerAuthHeader(token, headers=None):
    headers = headers or {}
    headers['Authorization'] = f'Bearer {token}'
    return headers


class HTMLFormParser(HTMLParser):
    def __init__(self, form_id):
        super().__init__()
        self._form_id = form_id
        self._inside_form = False
        self.target = None
        self.data = {}

    def _get_attr(self, attrs, name):
        for attr in attrs:
            if attr[0] == name:
                return attr[1]
        return None

    def handle_starttag(self, tag, attrs):
        if self._inside_form and tag == 'input':
            self.handle_input(attrs)
            return

        if tag == 'form' and self._get_attr(attrs, 'id') == self._form_id:
            self._inside_form = True
            self.target = self._get_attr(attrs, 'action')

    def handle_endtag(self, tag):
        if tag == 'form' and self._inside_form:
            self._inside_form = False

    def handle_input(self, attrs):
        if not self._inside_form:
            return

        name = self._get_attr(attrs, 'name')
        value = self._get_attr(attrs, 'value')

        if name:
            self.data[name] = value


class ScriptFormParser(HTMLParser):
    fields = []
    targetField = ''

    def __init__(self):
        super().__init__()
        self._inside_script = False
        self.data = {}
        self.target = None

    def handle_starttag(self, tag, attrs):
        if not self._inside_script and tag == 'script':
            self._inside_script = True

    def handle_endtag(self, tag):
        if self._inside_script and tag == 'script':
            self._inside_script = False

    def handle_data(self, data):
        if not self._inside_script:
            return

        match = re.search(r'templateModel: (.*?),\n', data)
        if not match:
            return

        result = json.loads(match.group(1))
        self.target = result.get(self.targetField, None)
        self.data = {k: v for k, v in result.items() if k in self.fields}

        match2 = re.search(r'csrf_token: \'(.*?)\'', data)
        if match2:
            self.data['_csrf'] = match2.group(1)


class CredentialsFormParser(ScriptFormParser):
    fields = ['relayState', 'hmac', 'registerCredentialsPath', 'error', 'errorCode']
    targetField = 'postAction'


class TermsAndConditionsFormParser(ScriptFormParser):
    fields = ['relayState', 'hmac', 'countryOfResidence', 'legalDocuments']
    targetField = 'loginUrl'

    def handle_data(self, data):
        if not self._inside_script:
            return

        super().handle_data(data)

        if 'countryOfResidence' in self.data:
            self.data['countryOfResidence'] = self.data['countryOfResidence'].upper()

        if 'legalDocuments' not in self.data:
            return

        for key in self.data['legalDocuments'][0]:
            # Skip unnecessary keys
            if key in ('skipLink', 'declineLink', 'majorVersion', 'minorVersion', 'changeSummary'):
                continue

            # Move values under a new key while converting boolean values to 'yes' or 'no'
            v = self.data['legalDocuments'][0][key]
            self.data[f'legalDocuments[0].{key}'] = ('yes' if v else 'no') if isinstance(v, bool) else v

        # Remove the original object
        del self.data['legalDocuments']



import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus
from weconnect.util import kelvinToCelsius

LOG = logging.getLogger("weconnect")


class TemperatureOutsideStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.temperatureOutside_K = AddressableAttribute(localAddress='temperatureOutside_K', parent=self, value=None, valueType=float)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update outside temperature status from dict')

        if 'value' in fromDict:
            self.temperatureOutside_K.fromDict(fromDict['value'], 'temperatureOutside_K')

        else:
            self.temperatureOutside_K.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + ['temperatureOutside_K']))

    def __str__(self):
        string = super().__str__()
        if self.temperatureOutside_K.enabled:
            string += f'\n\tOutside temperature {kelvinToCelsius(self.temperatureOutside_K.value)}°C'
        return string



from enum import IntEnum
import logging
from datetime import datetime

from weconnect.addressable import AddressableObject, AddressableAttribute

LOG = logging.getLogger("weconnect")


class GenericCapability(AddressableObject):
    def __init__(
        self,
        capabilityId,
        parent,
        fromDict=None,
        fixAPI=True,
    ):
        self.fixAPI = fixAPI
        super().__init__(localAddress=capabilityId, parent=parent)
        self.id = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=str)
        self.status = AddressableAttribute(localAddress='status', parent=self, value=None, valueType=list)
        self.expirationDate = AddressableAttribute(
            localAddress='expirationDate', parent=self, value=None, valueType=datetime)
        self.userDisablingAllowed = AddressableAttribute(
            localAddress='userDisablingAllowed', parent=self, value=None, valueType=bool)
        LOG.debug('Create capability from dict')
        if fromDict is not None:
            self.update(fromDict=fromDict)

    def update(self, fromDict):
        LOG.debug('Update capability from dict')

        self.id.fromDict(fromDict, 'id')

        if 'status' in fromDict:
            statuses = []
            for status in fromDict['status']:
                try:
                    statuses.append(GenericCapability.Status(status))
                except ValueError:
                    statuses.append(GenericCapability.Status(GenericCapability.Status.UNKNOWN))
                    LOG.debug('An unsupported status: %s was provided, please report this as a bug', status)
            self.status.setValueWithCarTime(statuses, lastUpdateFromCar=None, fromServer=True)
        else:
            self.status.enabled = False

        self.expirationDate.fromDict(fromDict, 'expirationDate')
        self.userDisablingAllowed.fromDict(fromDict, 'userDisablingAllowed')

        for key, value in {key: value for key, value in fromDict.items()
                           if key not in ['id', 'status', 'expirationDate', 'userDisablingAllowed']}.items():
            LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

    def __str__(self):
        returnString = f'[{self.id.value}]'
        if self.status.enabled and self.status.value is not None:
            returnString += f' Status: {", ".join([status.name for status in self.status.value])}'
        returnString += f' disabling: {self.userDisablingAllowed.value}'
        if self.expirationDate.enabled:
            returnString += f' (expires {self.expirationDate.value.isoformat()})'  # pylint: disable=no-member
        return returnString

    class Status(IntEnum):
        UNKNOWN = 0
        DEACTIVATED = 1001
        INITIALLY_DISABLED = 1003
        DISABLED_BY_USER = 1004
        OFFLINE_MODE = 1005
        WORKSHOP_MODE = 1006
        MISSING_OPERATION = 1007
        MISSING_SERVICE = 1008
        PLAY_PROTECTION = 1009
        POWER_BUDGET_REACHED = 1010
        DEEP_SLEEP = 1011
        LOCATION_DATA_DISABLED = 1013
        LICENSE_INACTIVE = 2001
        LICENSE_EXPIRED = 2002
        MISSING_LICENSE = 2003
        USER_NOT_VERIFIED = 3001
        TERMS_AND_CONDITIONS_NOT_ACCEPTED = 3002
        INSUFFICIENT_RIGHTS = 3003
        CONSENT_MISSING = 3004
        LIMITED_FEATURE = 3005
        AUTH_APP_CERT_ERROR = 3006
        STATUS_UNSUPPORTED = 4001



class RetrievalError(Exception):
    pass


class TooManyRequestsError(RetrievalError):
    pass


class SetterError(Exception):
    pass


class ControlError(SetterError):
    pass


class AuthentificationError(Exception):
    pass


class TemporaryAuthentificationError(AuthentificationError):
    pass


class APICompatibilityError(Exception):
    pass


class APIError(Exception):
    pass



import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class OdometerMeasurement(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.odometer = AddressableAttribute(
            localAddress='odometer', parent=self, value=None, valueType=int)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Odometer measurement from dict')

        if 'value' in fromDict:
            if 'odometer' in fromDict['value']:
                odometer = int(fromDict['value']['odometer'])
                if self.fixAPI and odometer == 0x7FFFFFFF:
                    odometer = None
                    LOG.info('%s: Attribute odometer was error value 0x7FFFFFFF. Setting error state instead'
                             ' of 2147483647 km.', self.getGlobalAddress())
                self.odometer.setValueWithCarTime(odometer, lastUpdateFromCar=None, fromServer=True)
            else:
                self.odometer.enabled = False
        else:
            self.odometer.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(
            ignoreAttributes + ['odometer']))

    def __str__(self):
        string = super().__str__()
        if self.odometer.enabled:
            string += f'\n\tOdometer: {self.odometer.value}km'
        return string



from datetime import datetime
import logging

from weconnect.addressable import AddressableObject, AddressableAttribute, AddressableDict

from weconnect.util import robustTimeParse, toBool

LOG = logging.getLogger("weconnect")


class Timer(AddressableObject):
    def __init__(
        self,
        parent,
        fromDict=None,
    ):
        super().__init__(localAddress=None, parent=parent)
        self.timerEnabled = AddressableAttribute(localAddress='enabled', parent=self, value=None, valueType=bool)
        self.climatisation = AddressableAttribute(localAddress='climatisation', parent=self, value=None, valueType=bool)
        self.recurringTimer = None
        self.singleTimer = None
        if fromDict is not None:
            self.update(fromDict)

    def update(self, fromDict):  # noqa: C901
        LOG.debug('Update timer from dict')

        if 'id' in fromDict:
            self.id = fromDict['id']
            self.localAddress = str(self.id)
        else:
            LOG.error('Timer is missing id attribute')

        if 'enabled' in fromDict:
            self.timerEnabled.setValueWithCarTime(toBool(fromDict['enabled']), lastUpdateFromCar=None, fromServer=True)
        else:
            self.timerEnabled.enabled = False

        if 'climatisation' in fromDict:
            self.climatisation.setValueWithCarTime(toBool(fromDict['climatisation']), lastUpdateFromCar=None, fromServer=True)
        else:
            self.timerEnabled.enabled = False

        if 'recurringTimer' in fromDict:
            if self.recurringTimer is None:
                self.recurringTimer = Timer.RecurringTimer(
                    localAddress='recurringTimer', parent=self, fromDict=fromDict['recurringTimer'])
            else:
                self.recurringTimer.update(fromDict=fromDict['recurringTimer'])
        elif self.recurringTimer is not None:
            self.recurringTimer.enabled = False
            self.recurringTimer = None

        if 'singleTimer' in fromDict:
            if self.singleTimer is None:
                self.singleTimer = Timer.SingleTimer(
                    localAddress='singleTimer', parent=self, fromDict=fromDict['singleTimer'])
            else:
                self.singleTimer.update(fromDict=fromDict['singleTimer'])
        elif self.singleTimer is not None:
            self.singleTimer.enabled = False
            self.singleTimer = None

        for key, value in {key: value for key, value in fromDict.items()
                           if key not in ['id', 'enabled', 'climatisation', 'recurringTimer', 'singleTimer']}.items():
            LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

    def __str__(self):
        string = f'{self.id}: Enabled: {self.timerEnabled.value}'
        if self.climatisation.enabled:
            string += f' (Climatisation: {self.climatisation.value})'
        if self.recurringTimer is not None and self.recurringTimer.enabled:
            string += f' at {self.recurringTimer} '
        if self.singleTimer is not None and self.singleTimer.enabled:
            string += f' at {self.singleTimer} '
        return string

    class RecurringTimer(AddressableObject):
        def __init__(
            self,
            localAddress,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=localAddress, parent=parent)
            self.startTime = AddressableAttribute(localAddress='startTime', parent=self, value=None, valueType=datetime)
            self.targetTime = AddressableAttribute(localAddress='targetTime', parent=self, value=None, valueType=datetime)
            self.recurringOn = AddressableDict(localAddress='recurringOn', parent=self)
            self.repetitionDays = AddressableDict(localAddress='repetitionDays', parent=self)
            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):  # noqa: C901
            LOG.debug('Update recurring timer from dict')

            if 'startTime' in fromDict:
                self.startTime.setValueWithCarTime(datetime.strptime(f'{fromDict["startTime"]}+00:00', '%H:%M%z'),
                                                   lastUpdateFromCar=None, fromServer=True)
            else:
                self.startTime.enabled = False

            if 'targetTime' in fromDict:
                self.targetTime.setValueWithCarTime(datetime.strptime(f'{fromDict["targetTime"]}+00:00', '%H:%M%z'),
                                                    lastUpdateFromCar=None, fromServer=True)
            else:
                self.targetTime.enabled = False

            if 'recurringOn' in fromDict and fromDict['recurringOn'] is not None:
                for day, state in fromDict['recurringOn'].items():
                    if day in self.recurringOn:
                        self.recurringOn[day].setValueWithCarTime(state, lastUpdateFromCar=None, fromServer=True)
                    else:
                        self.recurringOn[day] = AddressableAttribute(
                            localAddress=day, parent=self.recurringOn, value=state, valueType=bool)
                for day in [day for day in self.recurringOn.keys() if day not in fromDict['recurringOn'].keys()]:
                    del self.recurringOn[day]
            else:
                self.recurringOn.clear()
                self.recurringOn.enabled = False

            if 'repetitionDays' in fromDict and fromDict['repetitionDays'] is not None:
                for day in fromDict['repetitionDays']:
                    if day in self.repetitionDays:
                        self.repetitionDays[day].setValueWithCarTime(state, lastUpdateFromCar=None, fromServer=True)
                    else:
                        self.repetitionDays[day] = AddressableAttribute(
                            localAddress=day, parent=self.repetitionDays, value=state, valueType=bool)
                for day in [day for day in self.repetitionDays.keys() if day not in fromDict['repetitionDays']]:
                    del self.repetitionDays[day]
            else:
                self.repetitionDays.clear()
                self.repetitionDays.enabled = False

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['startTime', 'targetTime', 'recurringOn', 'repetitionDays']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            if self.startTime.enabled:
                string = f'{self.startTime.value.strftime("%H:%M")} on '  # pylint: disable=no-member
            if self.targetTime.enabled:
                string = f'{self.startTime.value.strftime("%H:%M")} on '  # pylint: disable=no-member
            if len(self.recurringOn):
                for day, value in self.recurringOn.items():
                    if value:
                        string += day + ' '
            elif len(self.repetitionDays):
                for day in self.repetitionDays:
                    string += day + ' '
            return string

    class SingleTimer(AddressableObject):
        def __init__(
            self,
            localAddress,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=localAddress, parent=parent)
            self.startDateTime = AddressableAttribute(
                localAddress='startDateTime', parent=self, value=None, valueType=datetime)
            self.targetDateTime = AddressableAttribute(
                localAddress='targetDateTime', parent=self, value=None, valueType=datetime)
            self.startDateTimeLocal = AddressableAttribute(
                localAddress='startDateTimeLocal', parent=self, value=None, valueType=datetime)
            self.targetDateTimeLocal = AddressableAttribute(
                localAddress='targetDateTimeLocal', parent=self, value=None, valueType=datetime)
            self.occurringOn = AddressableAttribute(
                localAddress='occurringOn', parent=self, value=None, valueType=str)
            self.startTime = AddressableAttribute(
                localAddress='startTime', parent=self, value=None, valueType=datetime)
            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update recurring timer from dict')

            if 'startDateTime' in fromDict:
                self.startDateTime.setValueWithCarTime(robustTimeParse(fromDict["startDateTime"]), lastUpdateFromCar=None, fromServer=True)
            else:
                self.startDateTime.enabled = False

            if 'targetDateTime' in fromDict:
                self.targetDateTime.setValueWithCarTime(robustTimeParse(fromDict["targetDateTime"]), lastUpdateFromCar=None, fromServer=True)
            else:
                self.targetDateTime.enabled = False

            if 'startDateTimeLocal' in fromDict:
                self.startDateTimeLocal.setValueWithCarTime(robustTimeParse(fromDict["startDateTimeLocal"]), lastUpdateFromCar=None, fromServer=True)
            else:
                self.startDateTimeLocal.enabled = False

            if 'targetDateTimeLocal' in fromDict:
                self.targetDateTimeLocal.setValueWithCarTime(robustTimeParse(fromDict["targetDateTimeLocal"]), lastUpdateFromCar=None, fromServer=True)
            else:
                self.targetDateTimeLocal.enabled = False
            self.occurringOn.fromDict(fromDict, 'occurringOn')

            if 'startTime' in fromDict:
                self.startTime.setValueWithCarTime(datetime.strptime(f'{fromDict["startTime"]}+00:00', '%H:%M%z'),
                                                   lastUpdateFromCar=None, fromServer=True)
            else:
                self.startTime.enabled = False

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['startDateTime',
                                              'targetDateTime',
                                              'startDateTimeLocal',
                                              'targetDateTimeLocal',
                                              'occurringOn',
                                              'startTime']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnString = ""
            if self.startDateTime.enabled:
                returnString += self.startDateTime.value.isoformat()  # pylint: disable=no-member
            elif self.startDateTimeLocal.enabled:
                returnString += self.startDateTimeLocal.value.isoformat()  # pylint: disable=no-member
            if self.targetDateTime.enabled:
                returnString += self.targetDateTime.value.isoformat()  # pylint: disable=no-member
            elif self.targetDateTimeLocal.enabled:
                returnString += self.targetDateTimeLocal.value.isoformat()  # pylint: disable=no-member
            if self.occurringOn.enabled:
                returnString += self.occurringOn.value
            if self.startTime.enabled:
                returnString += f' {self.startTime.value.strftime("%H:%M")}'
            return returnString



from enum import Enum
import logging

from weconnect.addressable import AddressableObject, AddressableAttribute
from weconnect.elements.generic_status import GenericStatus
from weconnect.util import toBool

LOG = logging.getLogger("weconnect")


class ReadinessStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.connectionState = None
        self.connectionWarning = None
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):  # noqa: C901
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update readiness status from dict')

        if 'value' in fromDict:
            if 'connectionState' in fromDict['value']:
                if self.connectionState is None:
                    self.connectionState = ReadinessStatus.ConnectionState(parent=self, fromDict=fromDict['value']['connectionState'])
                else:
                    self.connectionState.update(fromDict=fromDict['value']['connectionState'])
            elif self.connectionState is not None:
                self.connectionState.enabled = False
                self.connectionState = None

            if 'connectionWarning' in fromDict['value']:
                if self.connectionWarning is None:
                    self.connectionWarning = ReadinessStatus.ConnectionWarning(parent=self, fromDict=fromDict['value']['connectionWarning'])
                else:
                    self.connectionWarning.update(fromDict=fromDict['value']['connectionWarning'])
            elif self.connectionWarning is not None:
                self.connectionWarning.enabled = False
                self.connectionWarning = None

            super().update(fromDict=fromDict['value'], ignoreAttributes=(ignoreAttributes + ['connectionState', 'connectionWarning']))

    def __str__(self):
        string = super().__str__()
        if self.connectionState is not None and self.connectionState.enabled:
            string += f'\n\t{self.connectionState} '
        if self.connectionWarning is not None and self.connectionWarning.enabled:
            string += f'\n\t{self.connectionWarning} '
        return string

    class ConnectionState(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress='connectionState', parent=parent)
            self.isOnline = AddressableAttribute(localAddress='isOnline', parent=self, value=None, valueType=bool)
            self.isActive = AddressableAttribute(localAddress='isActive', parent=self, value=None, valueType=bool)
            self.batteryPowerLevel = AddressableAttribute(localAddress='batteryPowerLevel', parent=self, value=None,
                                                          valueType=ReadinessStatus.ConnectionState.BatteryPowerLevel)
            self.dailyPowerBudgetAvailable = AddressableAttribute(localAddress='dailyPowerBudgetAvailable', parent=self, value=None, valueType=bool)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update timer from dict')

            self.isOnline.fromDict(fromDict, 'isOnline')
            self.isActive.fromDict(fromDict, 'isActive')
            self.batteryPowerLevel.fromDict(fromDict, 'batteryPowerLevel')
            self.dailyPowerBudgetAvailable.fromDict(fromDict, 'dailyPowerBudgetAvailable')

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['isOnline', 'isActive', 'batteryPowerLevel', 'dailyPowerBudgetAvailable']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnString = 'Connection State: '
            if self.isOnline.enabled:
                returnString += f'\n\t\tIs online: {self.isOnline.value}'
            if self.isActive.enabled:
                returnString += f'\n\t\tIs active: {self.isActive.value}'
            if self.batteryPowerLevel.enabled:
                returnString += f'\n\t\tBattery power level: {self.batteryPowerLevel.value.value}'  # pylint: disable=no-member
            if self.dailyPowerBudgetAvailable.enabled:
                returnString += f'\n\t\tDaily Power Budget Available: {self.dailyPowerBudgetAvailable.value}'
            return returnString

        class BatteryPowerLevel(Enum):
            ACCESS = 'access'
            COMFORT = 'comfort'
            EMERGENCY = 'emergency'
            LOCATION = 'location'
            OFF = 'off'
            UNKNOWN = 'unknown battery power level'

    class ConnectionWarning(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress='connectionWarning', parent=parent)
            self.insufficientBatteryLevelWarning = AddressableAttribute(localAddress='insufficientBatteryLevelWarning', parent=self, value=None, valueType=bool)
            self.dailyPowerBudgetWarning = AddressableAttribute(localAddress='dailyPowerBudgetWarning', parent=self, value=None, valueType=bool)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update timer from dict')

            if 'insufficientBatteryLevelWarning' in fromDict:
                self.insufficientBatteryLevelWarning.setValueWithCarTime(toBool(fromDict['insufficientBatteryLevelWarning']), lastUpdateFromCar=None,
                                                                         fromServer=True)
            else:
                self.insufficientBatteryLevelWarning.enabled = False

            if 'dailyPowerBudgetWarning' in fromDict:
                self.dailyPowerBudgetWarning.setValueWithCarTime(toBool(fromDict['dailyPowerBudgetWarning']), lastUpdateFromCar=None, fromServer=True)
            else:
                self.dailyPowerBudgetWarning.enabled = False

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['insufficientBatteryLevelWarning', 'dailyPowerBudgetWarning']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnString = 'Connection Warning: '
            if self.insufficientBatteryLevelWarning.enabled:
                returnString += f'\n\t\tInsufficient Battery Level Warning: {self.insufficientBatteryLevelWarning.value}'
            if self.dailyPowerBudgetWarning.enabled:
                returnString += f'\n\t\tDaily Power Budget Warning: {self.dailyPowerBudgetWarning.value}'
            return returnString



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class ChargingStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.remainingChargingTimeToComplete_min = AddressableAttribute(
            localAddress='remainingChargingTimeToComplete_min', parent=self, value=None, valueType=int)
        self.chargingState = AddressableAttribute(
            localAddress='chargingState', value=None, parent=self, valueType=ChargingStatus.ChargingState)
        self.chargeMode = AddressableAttribute(
            localAddress='chargeMode', value=None, parent=self, valueType=ChargingStatus.ChargeMode)
        self.chargePower_kW = AddressableAttribute(
            localAddress='chargePower_kW', value=None, parent=self, valueType=float)
        self.chargeRate_kmph = AddressableAttribute(
            localAddress='chargeRate_kmph', value=None, parent=self, valueType=float)
        self.chargeType = AddressableAttribute(localAddress='chargeType', value=None, parent=self, valueType=ChargingStatus.ChargeType)
        self.chargingSettings = AddressableAttribute(localAddress='chargingSettings', value=None, parent=self, valueType=str)
        self.chargingScenario = AddressableAttribute(localAddress='chargingScenario', value=None, parent=self, valueType=ChargingStatus.ChargingScenario)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):  # noqa: C901
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update Charging status from dict')

        if 'value' in fromDict:
            self.remainingChargingTimeToComplete_min.fromDict(fromDict['value'], 'remainingChargingTimeToComplete_min')
            self.chargingState.fromDict(fromDict['value'], 'chargingState')
            self.chargeMode.fromDict(fromDict['value'], 'chargeMode')
            self.chargePower_kW.fromDict(fromDict['value'], 'chargePower_kW')
            if 'chargePower_kW' in fromDict['value']:
                chargePower_kW = float(fromDict['value']['chargePower_kW'])
                if self.fixAPI and chargePower_kW != 0 \
                        and self.chargingState.value in [ChargingStatus.ChargingState.OFF,
                                                         ChargingStatus.ChargingState.READY_FOR_CHARGING,
                                                         ChargingStatus.ChargingState.NOT_READY_FOR_CHARGING,
                                                         ChargingStatus.ChargingState.CHARGE_PURPOSE_REACHED_NOT_CONSERVATION_CHARGING,
                                                         ChargingStatus.ChargingState.ERROR]:
                    chargePower_kW = 0.0
                    LOG.debug('%s: Attribute chargePower_kW is %s while chargingState is %s. Setting 0 instead',
                              self.getGlobalAddress(), fromDict['value']['chargePower_kW'], self.chargingState.value)
                self.chargePower_kW.setValueWithCarTime(chargePower_kW, lastUpdateFromCar=None, fromServer=True)
            else:
                self.chargePower_kW.enabled = False
            if 'chargeRate_kmph' in fromDict['value']:
                chargeRate_kmph = float(fromDict['value']['chargeRate_kmph'])
                if self.fixAPI and chargeRate_kmph != 0 \
                        and self.chargingState.value in [ChargingStatus.ChargingState.OFF,
                                                         ChargingStatus.ChargingState.READY_FOR_CHARGING,
                                                         ChargingStatus.ChargingState.NOT_READY_FOR_CHARGING,
                                                         ChargingStatus.ChargingState.CHARGE_PURPOSE_REACHED_NOT_CONSERVATION_CHARGING,
                                                         ChargingStatus.ChargingState.ERROR]:
                    chargeRate_kmph = 0.0
                    LOG.debug('%s: Attribute chargeRate_kmph is %s while chargingState is %s. Setting 0 instead',
                              self.getGlobalAddress(), fromDict['value']['chargeRate_kmph'], self.chargingState.value)
                self.chargeRate_kmph.setValueWithCarTime(chargeRate_kmph, lastUpdateFromCar=None, fromServer=True)
            else:
                self.chargeRate_kmph.enabled = False
            self.chargeType.fromDict(fromDict['value'], 'chargeType')
            self.chargingSettings.fromDict(fromDict['value'], 'chargingSettings')
            self.chargingScenario.fromDict(fromDict['value'], 'chargingScenario')
        else:
            self.remainingChargingTimeToComplete_min.enabled = False
            self.chargingState.enabled = False
            self.chargeMode.enabled = False
            self.chargePower_kW.enabled = False
            self.chargeRate_kmph.enabled = False
            self.chargeType.enabled = False
            self.chargingSettings.enabled = False
            self.chargingScenario.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + [
                                                                'remainingChargingTimeToComplete_min',
                                                                'chargingState',
                                                                'chargeMode',
                                                                'chargePower_kW',
                                                                'chargeRate_kmph',
                                                                'chargeType',
                                                                'chargingSettings',
                                                                'chargingScenario'
                                                            ]))

    def __str__(self):
        string = super().__str__()
        if self.chargingState.enabled:
            string += f'\n\tState: {self.chargingState.value.value}'  # pylint: disable=no-member
        if self.chargeMode.enabled:
            string += f'\n\tMode: {self.chargeMode.value.value}'  # pylint: disable=no-member
        if self.remainingChargingTimeToComplete_min.enabled:
            string += f'\n\tRemaining Charging Time: {self.remainingChargingTimeToComplete_min.value} minutes'
        if self.chargePower_kW.enabled:
            string += f'\n\tCharge Power: {self.chargePower_kW.value} kW'
        if self.chargeRate_kmph.enabled:
            string += f'\n\tCharge Rate: {self.chargeRate_kmph.value} km/h'
        if self.chargeType.enabled:
            string += f'\n\tCharge Type: {self.chargeType.value.value}'
        if self.chargingSettings.enabled:
            string += f'\n\tCharging Settings: {self.chargingSettings.value}'
        if self.chargingScenario.enabled:
            string += f'\n\tCharging Scenario: {self.chargingScenario.value}'
        return string

    class ChargingState(Enum,):
        OFF = 'off'
        READY_FOR_CHARGING = 'readyForCharging'
        NOT_READY_FOR_CHARGING = 'notReadyForCharging'
        CONSERVATION = 'conservation'
        CHARGE_PURPOSE_REACHED_NOT_CONSERVATION_CHARGING = 'chargePurposeReachedAndNotConservationCharging'
        CHARGE_PURPOSE_REACHED_CONSERVATION = 'chargePurposeReachedAndConservation'
        CHARGING = 'charging'
        ERROR = 'error'
        UNSUPPORTED = 'unsupported'
        DISCHARGING = 'discharging'
        UNKNOWN = 'unknown charging state'

    class ChargeMode(Enum,):
        MANUAL = 'manual'
        INVALID = 'invalid'
        OFF = 'off'
        TIMER = 'timer'
        ONLY_OWN_CURRENT = 'onlyOwnCurrent'
        PREFERRED_CHARGING_TIMES = 'preferredChargingTimes'
        TIMER_CHARGING_WITH_CLIMATISATION = 'timerChargingWithClimatisation'
        HOME_STORAGE_CHARGING = 'homeStorageCharging'
        IMMEDIATE_DISCHARGING = 'immediateDischarging'
        UNKNOWN = 'unknown charge mode'

    class ChargeType(Enum,):
        INVALID = 'invalid'
        OFF = 'off'
        AC = 'ac'
        DC = 'dc'
        UNSUPPORTED = 'unsupported'
        UNKNOWN = 'unknown charge type'

    class ChargingScenario(Enum,):
        OFF = 'off'
        INITIALIZATION_CHARGING_COMMUNICATION = 'initializationChargingCommunication'
        IMMEDIATELY_CHARGING_ACTIVE = 'immediatelyChargingActive'
        IMMEDIATELY_CHARGING_FINISHED = 'immediatelyChargingFinished'
        CHARGING_TO_DEPARTURE_TIME_WAITING = 'chargingToDepartureTimeWaiting'
        CHARGING_TO_DEPARTURE_TIME_ACTIVE = 'chargingToDepartureTimeActive'
        OPTIMISED_CHARGING_FINISHED = 'optimisedChargingFinished'
        ERROR_CHARGING_SYSTEM = 'errorChargingSystem'
        UNKNOWN = 'unknown charging scenario'



from typing import TYPE_CHECKING, Optional, Dict, Any

import logging

from datetime import datetime

from weconnect.addressable import AddressableObject, AddressableAttribute
from weconnect.elements.enums import SpinState

if TYPE_CHECKING:
    from weconnect.elements.generic_status import GenericStatus

LOG: logging.Logger = logging.getLogger("weconnect")


class Error(AddressableObject):
    def __init__(
        self,
        localAddress: str,
        parent: Optional['GenericStatus'],
        fromDict: Dict[str, Any] = None,
    ) -> None:
        super().__init__(localAddress=localAddress, parent=parent)
        self.code: AddressableAttribute[int] = AddressableAttribute(localAddress='code', parent=self, value=None, valueType=int)
        self.message: AddressableAttribute[str] = AddressableAttribute(localAddress='message', parent=self, value=None, valueType=str)
        self.group: AddressableAttribute[int] = AddressableAttribute(localAddress='group', parent=self, value=None, valueType=int)
        self.info: AddressableAttribute[str] = AddressableAttribute(localAddress='info', parent=self, value=None, valueType=str)
        self.timestamp: AddressableAttribute[str] = AddressableAttribute(localAddress='timestamp', parent=self, value=None, valueType=datetime)
        self.retry: AddressableAttribute[bool] = AddressableAttribute(localAddress='retry', parent=self, value=None, valueType=bool)
        self.remainingTries: AddressableAttribute[int] = AddressableAttribute(localAddress='remainingTries', parent=self, value=None, valueType=int)
        self.spinLockedWaitingTime: AddressableAttribute[int] = AddressableAttribute(localAddress='spinLockedWaitingTime', parent=self, value=None,
                                                                                     valueType=int)
        self.spinState: AddressableAttribute[int] = AddressableAttribute(localAddress='spinState', parent=self, value=None, valueType=SpinState)

        if fromDict is not None:
            self.update(fromDict)

    def reset(self) -> None:
        if self.code.enabled:
            self.code.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.code.enabled = False
        if self.message.enabled:
            self.message.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.message.enabled = False
        if self.group.enabled:
            self.group.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.group.enabled = False
        if self.info.enabled:
            self.info.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.info.enabled = False
        if self.retry.enabled:
            self.retry.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.retry.enabled = False
        if self.remainingTries.enabled:
            self.remainingTries.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.remainingTries.enabled = False
        if self.spinLockedWaitingTime.enabled:
            self.spinLockedWaitingTime.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.spinLockedWaitingTime.enabled = False
        if self.spinState.enabled:
            self.spinState.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.spinState.enabled = False
        if self.enabled:
            self.enabled = False

    def update(self, fromDict: Dict[str, Any]) -> None:
        LOG.debug('Update Status Error from dict')

        self.code.fromDict(fromDict, 'code')
        self.message.fromDict(fromDict, 'message')
        self.group.fromDict(fromDict, 'group')
        self.info.fromDict(fromDict, 'info')
        self.timestamp.fromDict(fromDict, 'errorTimeStamp')
        self.retry.fromDict(fromDict, 'retry')
        self.remainingTries.fromDict(fromDict, 'remainingTries')
        self.spinLockedWaitingTime.fromDict(fromDict, 'spinLockedWaitingTime')
        self.spinState.fromDict(fromDict, 'spinState')

        if not self.code.enabled and not self.message.enabled and not self.code.enabled and not self.info.enabled \
                and not self.retry.enabled:
            self.enabled = False
        else:
            self.enabled = True

        for key, value in {key: value for key, value in fromDict.items()
                           if key not in ['code', 'message', 'group', 'info', 'errorTimeStamp', 'retry', 'remainingTries', 'spinLockedWaitingTime',
                                          'spinState']}.items():
            LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

    def __str__(self) -> str:
        returnString = f'Error {self.code.value}: {self.message.value} \n\tinfo: {self.info.value} \n\ttimestamp: {self.timestamp.value}'
        if self.remainingTries.enabled or self.spinLockedWaitingTime.enabled or self.spinState.enabled:
            returnString += '\n\tS-PIN:'
            if self.remainingTries.enabled:
                returnString += f' Remaining tries: {self.remainingTries.value}'
            if self.spinLockedWaitingTime.enabled:
                returnString += f' Locked, waiting time: {self.spinLockedWaitingTime.value}'
            if self.spinState.enabled:
                returnString += f' S-PIN State: {self.spinState.value.value}'
        return returnString



from enum import Flag, auto


class ErrorEventType(Flag):
    HTTP = auto()
    TIMEOUT = auto()
    CONNECTION = auto()
    JSON = auto()
    ALL = HTTP | TIMEOUT | CONNECTION | JSON  # pylint: disable=unsupported-binary-operation



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class AuxiliaryHeatingStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.remainingClimatisationTime_min = AddressableAttribute(
            localAddress='remainingClimatisationTime_min', parent=self, value=None, valueType=int)
        self.climatisationState = AddressableAttribute(localAddress='climatisationState', value=None, parent=self,
                                                       valueType=AuxiliaryHeatingStatus.ClimatizationState)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update AuxiliaryHeating status from dict')

        if 'value' in fromDict:
            self.climatisationState.fromDict(fromDict['value'], 'climatisationState')
            if 'remainingClimatisationTime_min' in fromDict['value']:
                remainingTime = int(fromDict['value']['remainingClimatisationTime_min'])
                if self.fixAPI and remainingTime != 0 and self.climatisationState.value == AuxiliaryHeatingStatus.ClimatizationState.OFF:
                    remainingTime = 0
                    LOG.debug('%s: Attribute remainingClimatisationTime_min is %s while climatisationState is %s. Setting 0 instead',
                              self.getGlobalAddress(), fromDict['value']['remainingClimatisationTime_min'], self.climatisationState.value)
                self.remainingClimatisationTime_min.setValueWithCarTime(remainingTime, lastUpdateFromCar=None, fromServer=True)
            else:
                self.remainingClimatisationTime_min.enabled = False
        else:
            self.remainingClimatisationTime_min.enabled = False
            self.climatisationState.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(
            ignoreAttributes + ['remainingClimatisationTime_min', 'climatisationState']))

    def __str__(self):
        string = super().__str__()
        if self.climatisationState.enabled:
            string += f'\n\tState: {self.climatisationState.value.value}'  # pylint: disable=no-member
        if self.remainingClimatisationTime_min.enabled:
            string += f'\n\tRemaining Climatization Time: {self.remainingClimatisationTime_min.value} min'
        return string

    class ClimatizationState(Enum,):
        OFF = 'off'
        HEATING = 'heating'
        COOLING = 'cooling'
        VENTILATION = 'ventilation'
        INVALID = 'invalid'
        UNKNOWN = 'unknown climatization state'



from enum import Enum
import logging

from weconnect.addressable import AddressableObject, AddressableAttribute, AddressableDict
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class AccessStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.overallStatus = AddressableAttribute(localAddress='overallStatus', parent=self, value=None, valueType=AccessStatus.OverallState)
        self.doorLockStatus = AddressableAttribute(localAddress='doorLockStatus', parent=self, value=None, valueType=AccessStatus.Door.LockState)
        self.doors = AddressableDict(localAddress='doors', parent=self)
        self.windows = AddressableDict(localAddress='windows', parent=self)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):  # noqa: C901
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update access status from dict')

        if 'value' in fromDict:
            self.overallStatus.fromDict(fromDict['value'], 'overallStatus')

            self.doorLockStatus.fromDict(fromDict['value'], 'doorLockStatus')

            if 'doors' in fromDict['value'] and fromDict['value']['doors'] is not None:
                for doorDict in fromDict['value']['doors']:
                    if 'name' in doorDict:
                        if doorDict['name'] in self.doors:
                            self.doors[doorDict['name']].update(fromDict=doorDict)
                        else:
                            self.doors[doorDict['name']] = AccessStatus.Door(fromDict=doorDict, parent=self.doors)
                for doorName in [doorName for doorName in self.doors.keys()
                                 if doorName not in [door['name'] for door in fromDict['value']['doors'] if 'name' in door]]:
                    del self.doors[doorName]
            else:
                self.doors.clear()
                self.doors.enabled = False

            if 'windows' in fromDict['value'] and fromDict['value']['windows'] is not None:
                for windowDict in fromDict['value']['windows']:
                    if 'name' in windowDict:
                        if windowDict['name'] in self.windows:
                            self.windows[windowDict['name']].update(fromDict=windowDict)
                        else:
                            self.windows[windowDict['name']] = AccessStatus.Window(fromDict=windowDict, parent=self.windows)
                for windowName in [windowName for windowName in self.windows.keys()
                                   if windowName not in [window['name']
                                   for window in fromDict['value']['windows'] if 'name' in window]]:
                    del self.doors[windowName]
            else:
                self.windows.clear()
                self.windows.enabled = False
        else:
            self.overallStatus.enabled = False
            self.doors.clear()
            self.doors.enabled = False
            self.windows.clear()
            self.windows.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['overallStatus', 'doorLockStatus', 'doors', 'windows']))

    def __str__(self):
        string = super().__str__()
        if self.overallStatus is not None and self.overallStatus.enabled:
            string += f'\n\tOverall Status: {self.overallStatus.value.value}'
        if self.doorLockStatus is not None and self.doorLockStatus.enabled:
            string += f'\n\tDoor Lock Status: {self.doorLockStatus.value.value}'
        if len(self.doors) > 0:
            string += f'\n\tDoors: {len(self.doors)} items'
            for door in self.doors.values():
                string += f'\n\t\t{door}'
        if len(self.windows) > 0:
            string += f'\n\tWindows: {len(self.windows)} items'
            for window in self.windows.values():
                string += f'\n\t\t{window}'
        return string

    class OverallState(Enum):
        SAFE = 'safe'
        UNSAFE = 'unsafe'
        INVALID = 'invalid'
        UNKNOWN = 'unknown overall state'

    class Door(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=None, parent=parent)
            self.openState = AddressableAttribute(
                localAddress='openState', parent=self, value=None, valueType=AccessStatus.Door.OpenState)
            self.lockState = AddressableAttribute(
                localAddress='lockState', parent=self, value=None, valueType=AccessStatus.Door.LockState)
            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update door from dict')

            if 'name' in fromDict:
                self.id = fromDict['name']
                self.localAddress = self.id
            else:
                LOG.error('Door is missing name attribute')

            if 'status' in fromDict and fromDict['status']:
                if 'locked' in fromDict['status']:
                    self.lockState.setValueWithCarTime(
                        AccessStatus.Door.LockState.LOCKED, lastUpdateFromCar=None, fromServer=True)
                elif 'unlocked' in fromDict['status']:
                    self.lockState.setValueWithCarTime(
                        AccessStatus.Door.LockState.UNLOCKED, lastUpdateFromCar=None, fromServer=True)
                else:
                    self.lockState.setValueWithCarTime(
                        AccessStatus.Door.LockState.UNKNOWN, lastUpdateFromCar=None, fromServer=True)

                if 'open' in fromDict['status']:
                    self.openState.setValueWithCarTime(AccessStatus.Door.OpenState.OPEN,
                                                       lastUpdateFromCar=None, fromServer=True)
                elif 'closed' in fromDict['status']:
                    self.openState.setValueWithCarTime(
                        AccessStatus.Door.OpenState.CLOSED, lastUpdateFromCar=None, fromServer=True)
                elif 'unsupported' in fromDict['status']:
                    self.openState.setValueWithCarTime(
                        AccessStatus.Door.OpenState.UNSUPPORTED, lastUpdateFromCar=None, fromServer=True)
                elif 'invalid' in fromDict['status']:
                    self.openState.setValueWithCarTime(
                        AccessStatus.Door.OpenState.INVALID, lastUpdateFromCar=None, fromServer=True)
                else:
                    self.openState.setValueWithCarTime(
                        AccessStatus.Door.OpenState.UNKNOWN, lastUpdateFromCar=None, fromServer=True)
            else:
                self.lockState.enabled = False
                self.openState.enabled = False

            for key, value in {key: value for key, value in fromDict.items() if key not in ['name', 'status']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnString = f'{self.id}: '
            if self.openState.enabled:
                returnString += f'{self.openState.value.value}'  # pylint: disable=no-member
            if self.lockState.enabled:
                returnString += f', {self.lockState.value.value}'  # pylint: disable=no-member
            return returnString

        class OpenState(Enum):
            OPEN = 'open'
            CLOSED = 'closed'
            UNSUPPORTED = 'unsupported'
            INVALID = 'invalid'
            UNKNOWN = 'unknown open state'

        class LockState(Enum):
            LOCKED = 'locked'
            UNLOCKED = 'unlocked'
            INVALID = 'invalid'
            UNKNOWN = 'unknown lock state'

    class Window(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=None, parent=parent)
            self.openState = AddressableAttribute(
                localAddress='openState', parent=self, value=None, valueType=AccessStatus.Window.OpenState)
            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update window from dict')

            if 'name' in fromDict:
                self.id = fromDict['name']
                self.localAddress = self.id
            else:
                LOG.error('Window is missing name attribute')

            if 'status' in fromDict and fromDict['status']:
                if 'open' in fromDict['status']:
                    self.openState.setValueWithCarTime(
                        AccessStatus.Window.OpenState.OPEN, lastUpdateFromCar=None, fromServer=True)
                elif 'closed' in fromDict['status']:
                    self.openState.setValueWithCarTime(
                        AccessStatus.Window.OpenState.CLOSED, lastUpdateFromCar=None, fromServer=True)
                elif 'unsupported' in fromDict['status']:
                    self.openState.setValueWithCarTime(AccessStatus.Window.OpenState.UNSUPPORTED, lastUpdateFromCar=None)
                elif 'invalid' in fromDict['status']:
                    self.openState.setValueWithCarTime(AccessStatus.Window.OpenState.INVALID, lastUpdateFromCar=None)
                else:
                    self.openState.setValueWithCarTime(
                        AccessStatus.Window.OpenState.UNKNOWN, lastUpdateFromCar=None, fromServer=True)
                    LOG.warning('No unsupported window status: %s was provided, please report this as a bug', fromDict['status'])
            else:
                self.openState.enabled = False

            for key, value in {key: value for key, value in fromDict.items() if key not in ['name', 'status']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            return f'{self.id}: {self.openState.value.value}'  # pylint: disable=no-member

        class OpenState(Enum,):
            OPEN = 'open'
            CLOSED = 'closed'
            UNSUPPORTED = 'unsupported'
            INVALID = 'invalid'
            UNKNOWN = 'unknown open state'



from urllib3.util.retry import Retry


class BlacklistRetry(Retry):
    def __init__(self, status_blacklist=None, **kwargs):
        self.status_blacklist = status_blacklist
        super().__init__(**kwargs)

    def is_retry(self, method, status_code, has_retry_after=False):
        if self.status_blacklist is not None and status_code in self.status_blacklist:
            return False
        else:
            return super().is_retry(method, status_code, has_retry_after)



from enum import Enum


class ControlInputEnum(Enum):
    @classmethod
    def allowedValues(cls):
        return []


class ControlOperation(ControlInputEnum):
    START = 'start'
    STOP = 'stop'
    NONE = 'none'
    SETTINGS = 'settings'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [ControlOperation.START, ControlOperation.STOP]


class AccessControlOperation(ControlInputEnum):
    LOCK = 'lock'
    UNLOCK = 'unlock'
    NONE = 'none'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [AccessControlOperation.LOCK, AccessControlOperation.UNLOCK]


class HonkAndFlashControlOperation(ControlInputEnum):
    FLASH = 'flash'
    HONK_AND_FLASH = 'honkandflash'
    NONE = 'none'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [HonkAndFlashControlOperation.FLASH, HonkAndFlashControlOperation.HONK_AND_FLASH]


class Operation(Enum):
    START = 'start'
    STOP = 'stop'
    SETTINGS = 'settings'
    LOCK = 'lock'
    UNLOCK = 'unlock'
    FLASH = 'flash'
    HONK_AND_FLASH = 'honkandflash'
    TIMERS = 'timers'
    MDOE = 'mode'
    PROFILES = 'profiles'
    UNKNOWN = 'unknown'



from typing import Any, Dict
from urllib3.util.retry import Retry
from urllib.parse import parse_qsl, urlparse, urlsplit, urljoin

import logging

import requests
from requests.adapters import HTTPAdapter
from requests.models import CaseInsensitiveDict

from weconnect.auth.auth_util import CredentialsFormParser, HTMLFormParser, TermsAndConditionsFormParser
from weconnect.auth.openid_session import OpenIDSession
from weconnect.errors import APICompatibilityError, AuthentificationError, RetrievalError


LOG = logging.getLogger("weconnect")


class VWWebSession(OpenIDSession):
    def __init__(self, sessionuser, acceptTermsOnLogin=False, **kwargs):
        super(VWWebSession, self).__init__(**kwargs)
        self.sessionuser = sessionuser
        self.acceptTermsOnLogin = acceptTermsOnLogin

        # Set up the web session
        retries = Retry(
            total=self.retries,
            backoff_factor=0.1,
            status_forcelist=[500],
            raise_on_status=False
        )

        self.websession: requests.Session = requests.Session()
        self.websession.proxies.update(self.proxies)
        self.websession.mount('https://', HTTPAdapter(max_retries=retries))
        self.websession.headers = CaseInsensitiveDict({
            'user-agent': 'Volkswagen/3.51.1-android/14',
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,'
                      'application/signed-exchange;v=b3',
            'accept-language': 'en-US,en;q=0.9',
            'accept-encoding': 'gzip, deflate',
            'x-requested-with': 'de.volkswagen.carnet.eu.eremote',
            'x-android-package-name': 'com.volkswagen.weconnect',
            'upgrade-insecure-requests': '1',
        })

    def doWebAuth(self, url: str) -> str:
        # Check if we already have the final OAuth callback URL
        if url.startswith('weconnect://authenticated'):
            LOG.info("Already have OAuth callback URL with tokens, returning immediately")
            return url.replace('weconnect://authenticated#', 'https://egal?')
        
        # Get the login form
        emailForm = self._get_login_form(url)

        if emailForm:
            # Legacy authentication flow
            # Set email to the provided username
            emailForm.data['email'] = self.sessionuser.username

            # Get password form
            passwordForm = self._get_password_form(
                urljoin('https://identity.vwgroup.io', emailForm.target),
                emailForm.data
            )

            # Set credentials
            passwordForm.data['email'] = self.sessionuser.username
            passwordForm.data['password'] = self.sessionuser.password

            # Log in and get the redirect URL
            url = self._handle_login(
                f'https://identity.vwgroup.io/signin-service/v1/{self.client_id}/{passwordForm.target}',
                passwordForm.data
            )
        else:
            # New authentication flow
            url = self._handle_new_auth_flow(url)

        # Check URL for terms and conditions
        while True:
            LOG.debug(f"DEBUG [do_web_auth]: Processing URL in while loop: {url[:150]}")

            # Check for custom scheme FIRST, before any URL manipulation
            if url.startswith('weconnect://authenticated'):
                LOG.info(f"Reached final OAuth callback URL with tokens")
                LOG.debug(f"DEBUG [do_web_auth]: Found weconnect://authenticated URL, breaking loop to return transformed URL")
                break

            # Also check for redirect_uri if it's a custom scheme
            if self.redirect_uri and url.startswith(self.redirect_uri):
                LOG.info(f"Reached redirect URI: {self.redirect_uri}")
                LOG.debug(f"DEBUG [do_web_auth]: Found redirect_uri URL, breaking loop")
                break

            # Check for any other custom scheme to prevent HTTP requests
            if url.startswith('weconnect://'):
                LOG.info(f"Found custom scheme URL, treating as final URL")
                LOG.debug(f"DEBUG [do_web_auth]: Found weconnect:// URL (non-authenticated), returning with transformation: {url[:100]}")
                return url.replace('weconnect://', 'https://egal?')

            # Only join URLs that are not custom schemes
            url = urljoin('https://identity.vwgroup.io', url)

            if 'terms-and-conditions' in url:
                if self.acceptTermsOnLogin:
                    url = self._handle_consent_form(url)
                else:
                    raise AuthentificationError(f'It seems like you need to accept the terms and conditions. '
                                                f'Try to visit the URL "{url}" or log into smartphone app.')

            LOG.debug(f"DEBUG [do_web_auth]: Making HTTP GET request to: {url[:100]}")
            response = self.websession.get(url, allow_redirects=False)
            if response.status_code == requests.codes['internal_server_error']:
                raise RetrievalError('Temporary server error during login')

            if 'Location' not in response.headers:
                raise APICompatibilityError('Forwarding without Location in headers')

            url = response.headers['Location']
            LOG.debug(f"DEBUG [do_web_auth]: Got Location header, new URL: {url[:100]}")

            # Check again after getting new URL from Location header
            if url.startswith('weconnect://authenticated'):
                LOG.info(f"OAuth flow completed, received callback URL with tokens")
                LOG.debug(f"DEBUG [do_web_auth]: Found weconnect://authenticated in Location header, breaking loop")
                break

            # Check for ANY weconnect:// scheme to prevent invalid requests
            if url.startswith('weconnect://'):
                LOG.info(f"OAuth flow reached custom scheme URL")
                LOG.debug(f"DEBUG [do_web_auth]: Found weconnect:// in Location header, breaking loop: {url[:100]}")
                break

            # Also check for redirect_uri if it's set
            if self.redirect_uri and url.startswith(self.redirect_uri):
                LOG.info(f"OAuth flow completed, received redirect URI")
                LOG.debug(f"DEBUG [do_web_auth]: Found redirect_uri in Location header: {url[:100]}")
                break

        LOG.debug(f"DEBUG [do_web_auth]: Exited while loop, final URL before transformation: {url[:150]}")

        # Handle the transformation based on the URL pattern
        if url.startswith('weconnect://authenticated#'):
            # Transform weconnect://authenticated# to https://egal?
            transformed_url = url.replace('weconnect://authenticated#', 'https://egal?')
            LOG.debug(f"DEBUG [do_web_auth]: Transformed weconnect://authenticated# URL to: {transformed_url[:150]}")
            return transformed_url
        elif self.redirect_uri and url.startswith(self.redirect_uri + '#'):
            # Transform redirect_uri# to https://egal?
            transformed_url = url.replace(self.redirect_uri + '#', 'https://egal?')
            LOG.debug(f"DEBUG [do_web_auth]: Transformed redirect_uri# URL to: {transformed_url[:150]}")
            return transformed_url
        else:
            LOG.warning(f"DEBUG [do_web_auth]: URL doesn't match expected patterns, returning as-is: {url[:150]}")
            return url

    def _get_login_form(self, url: str) -> HTMLFormParser:
        # Check for custom URL schemes before making HTTP requests
        if url.startswith('weconnect://'):
            LOG.info(f"[_get_login_form] Custom scheme URL detected, skipping legacy auth flow")
            LOG.debug(f"DEBUG [_get_login_form]: Custom scheme URL detected, returning None: {url[:150]}")
            return None

        # Also check if URL contains tokens already (might be a callback URL)
        if '#access_token=' in url or '#code=' in url:
            LOG.info(f"[_get_login_form] URL already contains OAuth tokens, skipping legacy auth flow")
            LOG.debug(f"DEBUG [_get_login_form]: URL contains tokens, returning None: {url[:150]}")
            return None
        while True:
            LOG.debug(f"DEBUG [_get_login_form]: Attempting to fetch: {url[:100]}")

            # Check for custom URL schemes during redirect loop
            if url.startswith('weconnect://'):
                LOG.info(f"[_get_login_form] Reached OAuth callback URL during redirects: {url[:100]}")
                LOG.debug(f"DEBUG [_get_login_form]: Custom scheme URL reached in redirect loop, returning None")
                return None
            
            response = self.websession.get(url, allow_redirects=False)
            if response.status_code == requests.codes['ok']:
                break

            if response.status_code in (requests.codes['found'], requests.codes['see_other']):
                if 'Location' not in response.headers:
                    raise APICompatibilityError('Forwarding without Location in headers')

                # Resolve relative URL to absolute URL
                url = urljoin(url, response.headers['Location'])

                # Check if the new URL is a custom scheme URL
                if url.startswith('weconnect://'):
                    LOG.info(f"[_get_login_form] OAuth callback URL found in Location header: {url[:100]}")
                    LOG.debug(f"DEBUG [_get_login_form]: Custom scheme URL in Location header, returning None")
                    return None
                continue

            raise APICompatibilityError(f'Retrieving login page was not successful, '
                                        f'status code: {response.status_code}')

        # Find login form on page to obtain inputs
        emailForm = HTMLFormParser(form_id='emailPasswordForm')
        emailForm.feed(response.text)

        if not emailForm.target or not all(x in emailForm.data for x in ['_csrf', 'relayState', 'hmac', 'email']):
            # Return None to indicate legacy form not found, new flow should be used
            return None

        return emailForm

    def _get_password_form(self, url: str, data: Dict[str, Any]) -> CredentialsFormParser:
        response = self.websession.post(url, data=data, allow_redirects=True)
        if response.status_code != requests.codes['ok']:
            raise APICompatibilityError(f'Retrieving credentials page was not successful, '
                                        f'status code: {response.status_code}')

        # Find login form on page to obtain inputs
        credentialsForm = CredentialsFormParser()
        credentialsForm.feed(response.text)

        if not credentialsForm.target or not all(x in credentialsForm.data for x in ['relayState', 'hmac', '_csrf']):
            raise APICompatibilityError('Could not find all required input fields on credentials page')

        if credentialsForm.data.get('error', None) is not None:
            if credentialsForm.data['error'] == 'validator.email.invalid':
                raise AuthentificationError('Error during login, email invalid')
            raise AuthentificationError(f'Error during login: {credentialsForm.data["error"]}')

        if 'errorCode' in credentialsForm.data:
            raise AuthentificationError('Error during login, is the username correct?')

        if credentialsForm.data.get('registerCredentialsPath', None) == 'register':
            raise AuthentificationError(f'Error during login, account {self.sessionuser.username} does not exist')

        return credentialsForm

    def _handle_login(self, url: str, data: Dict[str, Any]) -> str:
        response: requests.Response = self.websession.post(url, data=data, allow_redirects=False)

        if response.status_code == requests.codes['internal_server_error']:
            raise RetrievalError('Temporary server error during login')

        if response.status_code not in (requests.codes['found'], requests.codes['see_other']):
            raise APICompatibilityError(f'Forwarding expected (status code 302), '
                                        f'but got status code {response.status_code}')

        if 'Location' not in response.headers:
            raise APICompatibilityError('Forwarding without Location in headers')

        # Parse parameters from forwarding url
        params: Dict[str, str] = dict(parse_qsl(urlsplit(response.headers['Location']).query))

        # Check for login error
        if 'error' in params and params['error']:
            errorMessages: Dict[str, str] = {
                'login.errors.password_invalid': 'Password is invalid',
                'login.error.throttled': 'Login throttled, probably too many wrong logins. You have to wait '
                                         'a few minutes until a new login attempt is possible'
            }

            raise AuthentificationError(errorMessages.get(params['error'], params['error']))

        # Check for user ID
        if 'userId' not in params or not params['userId']:
            if 'updated' in params and params['updated'] == 'dataprivacy':
                raise AuthentificationError('You have to login at myvolkswagen.de and accept the terms and conditions')
            raise APICompatibilityError('No user ID provided')

        self.userId = params['userId']  # pylint: disable=unused-private-member
        return response.headers['Location']

    def _handle_new_auth_flow(self, url: str) -> str:
        """
        Handle the new authentication flow when legacy form is not available.
        
        Args:
            url (str): The authorization URL.
            
        Returns:
            str: The redirect URL after successful authentication.
            
        Raises:
            AuthenticationError: If authentication fails.
            APICompatibilityError: If there are issues with the authentication flow.
        """
        import re

        # Check if we already have the OAuth callback URL
        if url.startswith('weconnect://authenticated'):
            LOG.info("OAuth callback URL already present in new auth flow, returning immediately")
            return url

        # Get the authorization page (manually follow redirects to avoid custom scheme issues)
        max_initial_redirects = 5
        while max_initial_redirects > 0:
            # Check for custom scheme before making request
            if url.startswith('weconnect://'):
                LOG.info(f"Found custom scheme URL during initial auth page fetch")
                return url

            response = self.websession.get(url, allow_redirects=False)

            if response.status_code == requests.codes['ok']:
                break

            if response.status_code in (requests.codes['found'], requests.codes['see_other']):
                if 'Location' not in response.headers:
                    raise APICompatibilityError('Forwarding without Location in headers')
                url = urljoin(url, response.headers['Location'])
                max_initial_redirects -= 1
                continue

            raise APICompatibilityError(f'Failed to fetch authorization page, status code: {response.status_code}')

        if max_initial_redirects == 0:
            raise APICompatibilityError('Too many redirects while fetching authorization page')

        # Extract state token using regex
        state_match = re.search(r'<input[^>]*name="state"[^>]*value="([^"]*)"', response.text)
        state = state_match.group(1) if state_match else None

        if not state:
            raise APICompatibilityError('Could not find state token in authorization page')

        # Create login form data
        login_form = {
            'username': self.sessionuser.username,
            'password': self.sessionuser.password,
            'state': state
        }

        # Post to login URL
        login_url = f'https://identity.vwgroup.io/u/login?state={state}'
        LOG.debug(f"Posting to login URL: {login_url}")
        response = self.websession.post(login_url, data=login_form, allow_redirects=False)

        if response.status_code not in (requests.codes['found'], requests.codes['see_other']):
            raise AuthenticationError(f'Login failed with status code: {response.status_code}')

        if 'Location' not in response.headers:
            raise APICompatibilityError('No Location header in login response')

        # Follow redirects to get the final URL with authorization code
        redirect_url = response.headers['Location']
        LOG.debug(f"DEBUG [_handle_new_auth_flow]: Starting redirect follow loop with URL: {redirect_url[:150]}")
        max_depth = 10
        while max_depth > 0:
            LOG.debug(f"DEBUG [_handle_new_auth_flow]: Loop iteration, max_depth={max_depth}, URL: {redirect_url[:150]}")

            # Check for custom scheme IMMEDIATELY before any processing
            if redirect_url.startswith('weconnect://authenticated'):
                LOG.info(f"Successfully reached OAuth callback URL with custom scheme")
                LOG.debug(f"DEBUG [_handle_new_auth_flow]: Found weconnect://authenticated URL with tokens, returning immediately")
                return redirect_url

            # Also check for any weconnect:// scheme
            if redirect_url.startswith('weconnect://'):
                LOG.info(f"Found weconnect:// custom scheme URL")
                LOG.debug(f"DEBUG [_handle_new_auth_flow]: Found weconnect:// URL, returning immediately: {redirect_url[:150]}")
                return redirect_url

            if max_depth == 0:
                raise APICompatibilityError('Too many redirects in new auth flow')

            # Only process non-custom scheme URLs
            redirect_url = urljoin('https://identity.vwgroup.io', redirect_url)
            LOG.debug(f"DEBUG [_handle_new_auth_flow]: URL after urljoin: {redirect_url[:150]}")

            LOG.debug(f"DEBUG [_handle_new_auth_flow]: Making HTTP GET request to: {redirect_url[:100]}")
            response = self.websession.get(redirect_url, allow_redirects=False)

            if response.status_code == requests.codes['internal_server_error']:
                raise RetrievalError('Temporary server error during new auth flow')

            if 'Location' not in response.headers:
                LOG.debug(f"DEBUG [_handle_new_auth_flow]: No Location header in response, status code: {response.status_code}")
                raise APICompatibilityError('No Location header in redirect')

            redirect_url = response.headers['Location']
            LOG.debug(f"DEBUG [_handle_new_auth_flow]: Got Location header: {redirect_url[:150]}")

            # Check again after getting new redirect URL
            if redirect_url.startswith('weconnect://authenticated'):
                LOG.info(f"Successfully reached OAuth callback URL with custom scheme after redirect")
                LOG.debug(f"DEBUG [_handle_new_auth_flow]: Found weconnect://authenticated URL after redirect, returning immediately")
                return redirect_url

            # Also check for any weconnect:// scheme
            if redirect_url.startswith('weconnect://'):
                LOG.info(f"Found weconnect:// custom scheme URL after redirect")
                LOG.debug(f"DEBUG [_handle_new_auth_flow]: Found weconnect:// URL after redirect, returning: {redirect_url[:150]}")
                return redirect_url

            max_depth -= 1

        LOG.debug(f"DEBUG [_handle_new_auth_flow]: Exiting redirect loop after max iterations, returning URL: {redirect_url[:150]}")
        return redirect_url

    def _handle_consent_form(self, url: str) -> str:
        response = self.websession.get(url, allow_redirects=False)
        if response.status_code == requests.codes['internal_server_error']:
            raise RetrievalError('Temporary server error during login')

        # Find form on page to obtain inputs
        tcForm = TermsAndConditionsFormParser()
        tcForm.feed(response.text)

        # Remove query from URL
        url = urlparse(response.url)._replace(query='').geturl()

        response = self.websession.post(url, data=tcForm.data, allow_redirects=False)
        if response.status_code == requests.codes['internal_server_error']:
            raise RetrievalError('Temporary server error during login')

        if response.status_code not in (requests.codes['found'], requests.codes['see_other']):
            raise APICompatibilityError('Forwarding expected (status code 302), '
                                        f'but got status code {response.status_code}')

        if 'Location' not in response.headers:
            raise APICompatibilityError('Forwarding without Location in headers')

        return response.headers['Location']



import logging

from weconnect.addressable import AddressableLeaf, ChangeableAttribute
from weconnect.elements.generic_settings import GenericSettings
from weconnect.elements.enums import BatteryCareMode

LOG = logging.getLogger("weconnect")


class ChargingCareSettings(GenericSettings):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.batteryCareMode = ChangeableAttribute(
            localAddress='batteryCareMode', parent=self, value=None, valueType=BatteryCareMode)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

        self.batteryCareMode.addObserver(self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                         priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update Charging Care settings from dict')

        if 'value' in fromDict:
            self.batteryCareMode.fromDict(fromDict['value'], 'batteryCareMode')
        else:
            self.batteryCareMode.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + [
                                                                'batteryCareMode',
                                                            ]))

    def __str__(self):
        string = super().__str__()
        if self.batteryCareMode.enabled:
            string += f'\n\tBattery Care Mode: {self.batteryCareMode.value.value}'  # pylint: disable=no-member # this is a fales positive
        return string



import logging
import requests

from oauthlib.common import add_params_to_uri
from oauthlib.oauth2 import InsecureTransportError, is_secure_transport

from requests.models import CaseInsensitiveDict

from weconnect.auth.openid_session import AccessType
from weconnect.auth.vw_web_session import VWWebSession
from weconnect.errors import AuthentificationError, RetrievalError, TemporaryAuthentificationError

LOG = logging.getLogger("weconnect")


class WeChargeSession(VWWebSession):
    def __init__(self, sessionuser, **kwargs):
        super(WeChargeSession, self).__init__(client_id='0fa5ae01-ebc0-4901-a2aa-4dd60572ea0e@apps_vw-dilab_com',
                                              refresh_url='https://identity.vwgroup.io/oidc/v1/token',
                                              scope='openid profile address email cars vin',
                                              redirect_uri='wecharge://authenticated',
                                              state=None,
                                              sessionuser=sessionuser,
                                              **kwargs)

        self.headers = CaseInsensitiveDict({
            'accept': '*/*',
            'content-type': 'application/json',
            'content-version': '1',
            'x-newrelic-id': 'VgAEWV9QDRAEXFlRAAYPUA==',
            'user-agent': 'WeConnect/3 CFNetwork/1327.0.4 Darwin/21.2.0',
            'accept-language': 'de-de',
        })

    @property
    def wcAccessToken(self):
        if self._token is not None and 'wc_access_token' in self._token:
            return self._token.get('wc_access_token')
        return None

    def login(self):
        super(WeChargeSession, self).login()
        authorizationUrl = self.authorizationUrl(url='https://identity.vwgroup.io/oidc/v1/authorize')
        response = self.doWebAuth(authorizationUrl)
        self.fetchTokens('https://wecharge.apps.emea.vwapps.io/user-identity/v1/identity/login',
                         authorization_response=response)

    def refresh(self):
        self.refreshTokens(
            'https://wecharge.apps.emea.vwapps.io/user-identity/v1/identity/login',
        )

    def fetchTokens(
        self,
        token_url,
        authorization_response=None,
        **kwargs
    ):
        self.parseFromFragment(authorization_response)

        if all(key in self.token for key in ('state', 'id_token', 'access_token', 'code')):
            loginHeadersForm: CaseInsensitiveDict = self.headers.copy()
            loginHeadersForm['accept'] = 'application/json'
            loginHeadersForm["x-api-key"] = "yabajourasW9N8sm+9F/oP=="

            urlParams = [(('redirect_uri', self.redirect_uri)),
                         (('code', self.token["code"]))]
            token_url = add_params_to_uri(token_url, urlParams)

            tokenResponse = self.get(token_url, headers=loginHeadersForm, allow_redirects=False, access_type=AccessType.ID)
            if tokenResponse.status_code != requests.codes['ok']:
                raise TemporaryAuthentificationError(f'Token could not be fetched due to temporary WeConnect failure: {tokenResponse.status_code}')

            self.parseFromBody(tokenResponse.text)

            return self.token

    def refreshTokens(
        self,
        token_url,
        refresh_token=None,
        auth=None,
        timeout=None,
        headers=None,
        verify=True,
        proxies=None,
        **kwargs
    ):
        LOG.info('Refreshing tokens')
        if not token_url:
            raise ValueError("No token endpoint set for auto_refresh.")

        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        refresh_token = refresh_token or self.refresh_token
        if refresh_token is None:
            raise ValueError("Missing refresh token.")

        if headers is None:
            headers = self.headers

        urlParams = [(('redirect_uri', self.redirect_uri)),
                     (('refresh_token', refresh_token))]

        token_url = add_params_to_uri(token_url, urlParams)

        refreshHeaders: CaseInsensitiveDict = headers.copy()
        refreshHeaders['accept'] = 'application/json'
        refreshHeaders["x-api-key"] = "yabajourasW9N8sm+9F/oP=="

        tokenResponse = self.get(
            token_url,
            auth=auth,
            timeout=timeout,
            headers=refreshHeaders,
            verify=verify,
            withhold_token=False,
            proxies=proxies,
            access_type=AccessType.REFRESH
        )
        if tokenResponse.status_code == requests.codes['unauthorized']:
            raise AuthentificationError('Refreshing tokens failed: Server requests new authorization')
        elif tokenResponse.status_code in (requests.codes['internal_server_error'], requests.codes['service_unavailable'], requests.codes['gateway_timeout']):
            raise TemporaryAuthentificationError('Token could not be refreshed due to temporary WeConnect failure: {tokenResponse.status_code}')
        elif tokenResponse.status_code == requests.codes['ok']:
            self.parseFromBody(tokenResponse.text)
            if "refresh_token" not in self.token:
                LOG.debug("No new refresh token given. Re-using old.")
                self.token["refresh_token"] = refresh_token
            return self.token
        else:
            raise RetrievalError(f'Status Code from WeConnect while refreshing tokens was: {tokenResponse.status_code}')

    def addToken(self, uri, body=None, headers=None, access_type=AccessType.ACCESS, **kwargs):
        headers = headers or {}
        uri, headers, body = super(WeChargeSession, self).addToken(uri, body=body, headers=headers, access_type=access_type, **kwargs)

        if access_type == AccessType.ACCESS:
            if not (self.wcAccessToken):
                raise ValueError("Missing wc access token.")
            headers['wc_access_token'] = self.wcAccessToken

        return (uri, headers, body)



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute, AddressableObject
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class RangeStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.carType = AddressableAttribute(localAddress='carType', parent=self,
                                            value=None, valueType=RangeStatus.CarType)
        self.primaryEngine = RangeStatus.Engine(localAddress='primaryEngine', parent=self)
        self.secondaryEngine = RangeStatus.Engine(localAddress='secondaryEngine', parent=self)
        self.totalRange_km = AddressableAttribute(localAddress='totalRange_km', parent=self, value=None, valueType=int)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update range status from dict')

        if 'value' in fromDict:
            self.carType.fromDict(fromDict['value'], 'carType')

            if 'primaryEngine' in fromDict['value']:
                self.primaryEngine.update(fromDict['value']['primaryEngine'])
            else:
                self.primaryEngine.enabled = False

            if 'secondaryEngine' in fromDict['value']:
                self.secondaryEngine.update(fromDict['value']['secondaryEngine'])
            else:
                self.secondaryEngine.enabled = False

            if 'totalRange_km' in fromDict['value'] and self.fixAPI \
                    and round((self.totalRange_km.value or 0) * 0.621371) == int(fromDict['value']['totalRange_km']) and self.totalRange_km.value != 0:
                LOG.info('%s: Attribute totalRange_km was miscalculated (miles/km) this is a bug in the API and the new value will not be used',
                         self.getGlobalAddress())
            else:
                self.totalRange_km.fromDict(fromDict['value'], 'totalRange_km')

        else:
            self.carType.enabled = False
            self.primaryEngine.enabled = False
            self.secondaryEngine.enabled = False
            self.totalRange_km.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + ['carType',
                                                               'primaryEngine',
                                                               'secondaryEngine',
                                                               'totalRange_km']))

    def __str__(self):
        string = super().__str__()
        if self.carType.enabled:
            string += f'\n\tCar Type: {self.carType.value.value}'  # pylint: disable=no-member
        if self.totalRange_km.enabled:
            string += f'\n\tTotal Range: {self.totalRange_km.value} km'
        if self.primaryEngine.enabled:
            string += f'\n\tPrimary Engine: {self.primaryEngine}'
        if self.secondaryEngine.enabled:
            string += f'\n\tSecondary Engine: {self.secondaryEngine}'
        return string

    class Engine(AddressableObject):
        def __init__(
            self,
            localAddress,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=localAddress, parent=parent)
            self.type = AddressableAttribute(localAddress='type', parent=self, value=None,
                                             valueType=RangeStatus.Engine.EngineType)
            self.currentSOC_pct = AddressableAttribute(
                localAddress='currentSOC_pct', parent=self, value=None, valueType=int)
            self.currentFuelLevel_pct = AddressableAttribute(
                localAddress='currentFuelLevel_pct', parent=self, value=None, valueType=int)
            self.remainingRange_km = AddressableAttribute(
                localAddress='remainingRange_km', parent=self, value=None, valueType=int)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update Engine from dict')

            self.type.fromDict(fromDict, 'type')
            self.currentFuelLevel_pct.fromDict(fromDict, 'currentFuelLevel_pct')

            if ('remainingRange_km' in fromDict and self.parent.fixAPI
                and round((self.remainingRange_km.value or 0) * 0.621371) == int(fromDict['remainingRange_km']) and self.remainingRange_km.value != 0
                    and self.currentSOC_pct.value == int(fromDict['currentSOC_pct'])):
                LOG.info('%s: Attribute remainingRange_km was miscalculated (miles/km) this is a bug in the API and the new value will not be used',
                         self.getGlobalAddress())
            else:
                self.remainingRange_km.fromDict(fromDict, 'remainingRange_km')

            self.currentSOC_pct.fromDict(fromDict, 'currentSOC_pct')

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['type', 'currentSOC_pct', 'currentFuelLevel_pct', 'remainingRange_km']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            string = ""
            if self.type.enabled:
                string += f"{self.type.value.value} "
            if self.currentFuelLevel_pct.enabled:
                string += f" Fuel Level: {self.currentFuelLevel_pct.value}%"
            elif self.currentSOC_pct.enabled:
                string += f" SoC: {self.currentSOC_pct.value}%"
            if self.remainingRange_km.enabled:
                string += f" ({self.remainingRange_km.value} km)"
            return string

        class EngineType(Enum,):
            GASOLINE = 'gasoline'
            ELECTRIC = 'electric'
            PETROL = 'petrol'
            DIESEL = 'diesel'
            CNG = 'cng'
            LPG = 'lpg'
            INVALID = 'invalid'
            UNKNOWN = 'unknown engine type'

    class CarType(Enum,):
        ELECTRIC = 'electric'
        HYBRID = 'hybrid'
        GASOLINE = 'gasoline'
        PETROL = 'petrol'
        DIESEL = 'diesel'
        CNG = 'cng'
        LPG = 'lpg'
        INVALID = 'invalid'
        UNKNOWN = 'unknown car type'



from typing import Dict

import json
import logging
import requests

from oauthlib.common import to_unicode
from oauthlib.oauth2 import InsecureTransportError
from oauthlib.oauth2 import is_secure_transport

from requests.models import CaseInsensitiveDict
from weconnect.auth.openid_session import AccessType

from weconnect.auth.vw_web_session import VWWebSession
from weconnect.errors import AuthentificationError, RetrievalError, TemporaryAuthentificationError


LOG = logging.getLogger("weconnect")


class MyCupraSession(VWWebSession):
    def __init__(self, sessionuser, **kwargs):
        super(MyCupraSession, self).__init__(client_id='3c756d46-f1ba-4d78-9f9a-cff0d5292d51@apps_vw-dilab_com',
                                             refresh_url='https://identity.vwgroup.io/oidc/v1/token',
                                             scope='openid profile nickname birthdate phone',
                                             redirect_uri='cupra://oauth-callback',
                                             state=None,
                                             sessionuser=sessionuser,
                                             **kwargs)

        self.headers = CaseInsensitiveDict({
            'accept': '*/*',
            'content-type': 'application/json',
            'user-agent': 'CUPRAApp%20-%20Store/20220503 CFNetwork/1333.0.4 Darwin/21.5.0',
            'accept-language': 'de-de',
            'accept-encoding': 'gzip, deflate, br'
        })

    def login(self):
        authorizationUrl = self.authorizationUrl(url='https://identity.vwgroup.io/oidc/v1/authorize')
        response = self.doWebAuth(authorizationUrl)
        self.fetchTokens('https://identity.vwgroup.io/oidc/v1/token',
                         authorization_response=response
                         )

    def refresh(self):
        self.refreshTokens(
            'https://identity.vwgroup.io/oidc/v1/token',
        )

    def fetchTokens(
        self,
        token_url,
        authorization_response=None,
        **kwargs
    ):

        self.parseFromFragment(authorization_response)

        if all(key in self.token for key in ('state', 'id_token', 'access_token', 'code')):
            body: Dict[str, str] = {
                'code': self.token['code'],
                'redirect_uri': self.redirect_uri,
                'client_id': self.client_id,
                'client_secret': 'eb8814e641c81a2640ad62eeccec11c98effc9bccd4269ab7af338b50a94b3a2',
                'grant_type': 'authorization_code',
                'state': self.token['state'],
                'id_token': self.token['id_token']
            }

            loginHeadersForm: CaseInsensitiveDict = self.headers
            loginHeadersForm['content-type'] = 'application/x-www-form-urlencoded; charset=utf-8'

            tokenResponse = self.post(token_url, headers=loginHeadersForm, data=body, allow_redirects=False, access_type=AccessType.NONE)
            if tokenResponse.status_code != requests.codes['ok']:
                print(tokenResponse.text)
                raise TemporaryAuthentificationError(f'Token could not be fetched due to temporary MyCupra failure: {tokenResponse.status_code}')
            token = self.parseFromBody(tokenResponse.text)

            return token

    def parseFromBody(self, token_response, state=None):
        try:
            token = json.loads(token_response)
        except json.decoder.JSONDecodeError:
            raise TemporaryAuthentificationError('Token could not be refreshed due to temporary MyCupra failure: json could not be decoded')
        if 'accessToken' in token:
            token['access_token'] = token.pop('accessToken')
        if 'idToken' in token:
            token['id_token'] = token.pop('idToken')
        if 'refreshToken' in token:
            token['refresh_token'] = token.pop('refreshToken')
        fixedTokenresponse = to_unicode(json.dumps(token)).encode("utf-8")
        return super(MyCupraSession, self).parseFromBody(token_response=fixedTokenresponse, state=state)

    def refreshTokens(
        self,
        token_url,
        refresh_token=None,
        auth=None,
        timeout=None,
        headers=None,
        verify=True,
        proxies=None,
        **kwargs
    ):
        LOG.info('Refreshing tokens')
        if not token_url:
            raise ValueError("No token endpoint set for auto_refresh.")

        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        refresh_token = refresh_token or self.refreshToken

        if headers is None:
            headers = self.headers

        body: Dict[str, str] = {
            'client_id': self.client_id,
            'client_secret': 'eb8814e641c81a2640ad62eeccec11c98effc9bccd4269ab7af338b50a94b3a2',
            'grant_type': 'refresh_token',
            'refresh_token': self.token['refresh_token']
        }

        headers['content-type'] = 'application/x-www-form-urlencoded; charset=utf-8'

        tokenResponse = self.post(
            token_url,
            data=body,
            auth=auth,
            timeout=timeout,
            headers=headers,
            verify=verify,
            withhold_token=False,
            proxies=proxies,
            access_type=AccessType.NONE
        )
        if tokenResponse.status_code == requests.codes['unauthorized']:
            raise AuthentificationError('Refreshing tokens failed: Server requests new authorization')
        elif tokenResponse.status_code in (requests.codes['internal_server_error'], requests.codes['service_unavailable'], requests.codes['gateway_timeout']):
            raise TemporaryAuthentificationError('Token could not be refreshed due to temporary MyCupra failure: {tokenResponse.status_code}')
        elif tokenResponse.status_code == requests.codes['ok']:
            self.parseFromBody(tokenResponse.text)
            if "refresh_token" not in self.token:
                LOG.debug("No new refresh token given. Re-using old.")
                self.token["refresh_token"] = refresh_token
            return self.token
        else:
            raise RetrievalError(f'Status Code from MyCupra while refreshing tokens was: {tokenResponse.status_code}')

    def request(
        self,
        method,
        url,
        data=None,
        headers=None,
        withhold_token=False,
        access_type=AccessType.ACCESS,
        token=None,
        timeout=None,
        **kwargs
    ):
        """Intercept all requests and add userId if present."""
        if not is_secure_transport(url):
            raise InsecureTransportError()
        if self.userId is not None:
            headers = headers or {}
            headers['user-id'] = self.userId

        return super(MyCupraSession, self).request(method, url, headers=headers, data=data, withhold_token=withhold_token, access_type=access_type, token=token,
                                                   timeout=timeout, **kwargs)



from enum import Enum

from weconnect.elements.control_operation import ControlInputEnum


class MaximumChargeCurrent(ControlInputEnum,):
    MAXIMUM = 'maximum'
    REDUCED = 'reduced'
    INVALID = 'invalid'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [MaximumChargeCurrent.MAXIMUM, MaximumChargeCurrent.REDUCED]


class UnlockPlugState(ControlInputEnum,):
    OFF = 'off'
    ON = 'on'
    PERMANENT = 'permanent'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [UnlockPlugState.OFF, UnlockPlugState.ON]


class BatteryCareMode(ControlInputEnum,):
    ACTIVATED = 'activated'
    DEACTIVATED = 'deactivated'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [BatteryCareMode.ACTIVATED, BatteryCareMode.DEACTIVATED]


class CarType(Enum,):
    ELECTRIC = 'electric'
    FUEL = 'fuel'
    HYBRID = 'hybrid'
    GASOLINE = 'gasoline'
    PETROL = 'petrol'
    DIESEL = 'diesel'
    CNG = 'cng'
    LPG = 'lpg'
    INVALID = 'invalid'
    UNKNOWN = 'unknown car type'


class EngineType(Enum,):
    GASOLINE = 'gasoline'
    ELECTRIC = 'electric'
    PETROL = 'petrol'
    DIESEL = 'diesel'
    CNG = 'cng'
    LPG = 'lpg'
    INVALID = 'invalid'
    UNKNOWN = 'unknown engine type'


class SpinState(Enum,):
    DEFINED = 'DEFINED'
    UNKNOWN = 'unknown spin state'


class TargetSOCReachable(Enum,):
    INVALID = 'invalid'
    CALCULATING = 'calculating'
    NOT_REACHABLE = 'notReachable'
    REACHABLE = 'reachable'
    UNKNOWN = 'unknown'



from __future__ import annotations
from typing import Dict, List, Set, Tuple, Callable, Any, Optional, Union

import os
from threading import Lock
import string
import locale
import logging
import json
from datetime import datetime, timedelta

import requests

from weconnect.auth.session_manager import SessionManager, Service, SessionUser
from weconnect.elements.vehicle import Vehicle
from weconnect.domain import Domain
from weconnect.elements.charging_station import ChargingStation
from weconnect.elements.general_controls import GeneralControls
from weconnect.addressable import AddressableLeaf, AddressableObject, AddressableDict
from weconnect.errors import RetrievalError, TooManyRequestsError
from weconnect.weconnect_errors import ErrorEventType
from weconnect.util import ExtendedEncoder

LOG = logging.getLogger("weconnect")


class WeConnect(AddressableObject):  # pylint: disable=too-many-instance-attributes, too-many-public-methods
    """Main class used to interact with WeConnect"""

    def __init__(  # noqa: C901 # pylint: disable=too-many-arguments
        self,
        username: str,
        password: str,
        spin: Union[str, bool] = None,
        tokenfile: Optional[str] = None,
        updateAfterLogin: bool = True,
        loginOnInit: bool = False,
        fixAPI: bool = True,
        proxy: Optional[str] = None,
        maxAge: Optional[int] = None,
        maxAgePictures: Optional[int] = None,
        updateCapabilities: bool = True,
        updatePictures: bool = True,
        numRetries: int = 3,
        timeout: bool = None,
        selective: Optional[list[Domain]] = None,
        forceReloginAfter: Optional[int] = None,
        acceptTermsOnLogin: Optional[bool] = False,
    ) -> None:
        """Initialize WeConnect interface. If loginOnInit is true the user will be tried to login.
           If loginOnInit is true also an initial fetch of data is performed.

        Args:
            username (str): Username used with WeConnect. This is your volkswagen user.
            password (str): Password used with WeConnect. This is your volkswagen password.
            tokenfile (str, optional): Optional file to read/write token from/to. Defaults to None.
            updateAfterLogin (bool, optional): Update data from WeConnect after logging in (If set to false, update needs to be called manually).
            Defaults to True.
            loginOnInit (bool, optional): Login after initialization (If set to false, login needs to be called manually). Defaults to True.
            fixAPI (bool, optional): Automatically fix known issues with the WeConnect responses. Defaults to True.
            proxy (str, optional): Set a proxy IP adress and port
            maxAge (int, optional): Maximum age of the cache before date is fetched again. None means no caching. Defaults to None.
            maxAgePictures (Optional[int], optional):  Maximum age of the pictures in the cache before date is fetched again. None means no caching.
            Defaults to None.
            updateCapabilities (bool, optional): Also update the information about the cars capabilities. Defaults to True.
            updatePictures (bool, optional):  Also fetch and update pictures. Defaults to True.
            numRetries (int, optional): Number of retries when http requests are failing. Defaults to 3.
            timeout (bool, optional, optional): Timeout in seconds used for http connections to the VW servers
            selective (list[Domain], optional): Domains to request data for
            forceReloginAfter (int, optional): Force a full relogin after number of seconds. This might be necessary to get fresh data
        """
        super().__init__(localAddress='', parent=None)
        self.lock = Lock()
        self.username: str = username
        self.password: str = password
        self.spin: Union[str, bool] = spin

        self.__session: requests.Session = requests.Session()

        self.__vehicles: AddressableDict[str, Vehicle] = AddressableDict(localAddress='vehicles', parent=self)
        self.__stations: AddressableDict[str, ChargingStation] = AddressableDict(localAddress='chargingStations', parent=self)
        self.__controls: GeneralControls = GeneralControls(localAddress='controls', parent=self)
        self.__cache: Dict[str, Any] = {}
        self.fixAPI: bool = fixAPI
        self.proxy: Optional[str] = proxy

        if proxy:
            self.proxystring = {'http': 'http://' + self.proxy, 'https': 'http://' + self.proxy}
        else:
            self.proxystring = ""

        self.maxAge: Optional[int] = maxAge
        self.maxAgePictures: Optional[int] = maxAgePictures
        self.latitude: Optional[float] = None
        self.longitude: Optional[float] = None
        self.searchRadius: Optional[int] = None
        self.market: Optional[str] = None
        self.useLocale: Optional[str] = locale.getlocale()[0]
        self.__elapsed: List[timedelta] = []

        self.__enableTracker: bool = False

        self.__errorObservers: Set[Tuple[Callable[[Optional[Any], ErrorEventType], None], ErrorEventType]] = set()

        self.tokenfile = tokenfile

        self.__manager = SessionManager(tokenstorefile=tokenfile)
        self.__session = self.__manager.getSession(Service.WE_CONNECT, SessionUser(username=username, password=password))
        self.__session.proxies.update(self.proxystring)
        self.__session.timeout = timeout
        self.__session.retries = numRetries
        self.__session.forceReloginAfter = forceReloginAfter
        self.__session.acceptTermsOnLogin = acceptTermsOnLogin

        if loginOnInit:
            self.__session.login()

        if updateAfterLogin:
            self.update(updateCapabilities=updateCapabilities, updatePictures=updatePictures, selective=selective)

    def __del__(self) -> None:
        self.disconnect()
        return super().__del__()

    def disconnect(self) -> None:
        pass

    @property
    def session(self) -> requests.Session:
        return self.__session

    @property
    def cache(self) -> Dict[str, Any]:
        return self.__cache

    def persistTokens(self) -> None:
        if self.__manager is not None and self.tokenfile is not None:
            self.__manager.saveTokenstore(self.tokenfile)

    def persistCacheAsJson(self, filename: str) -> None:
        with open(filename, 'w', encoding='utf8') as file:
            json.dump(self.__cache, file, cls=ExtendedEncoder)
        LOG.info('Writing cachefile %s', filename)

    def fillCacheFromJson(self, filename: str, maxAge: int, maxAgePictures: Optional[int] = None) -> None:
        self.maxAge = maxAge
        if maxAgePictures is None:
            self.maxAgePictures = maxAge
        else:
            self.maxAgePictures = maxAgePictures

        try:
            with open(filename, 'r', encoding='utf8') as file:
                self.__cache = json.load(file)
        except json.decoder.JSONDecodeError:
            LOG.error('Cachefile %s seems corrupted will delete it and try to create a new one. '
                      'If this problem persists please check if a problem with your disk exists.', filename)
            os.remove(filename)
        LOG.info('Reading cachefile %s', filename)

    def fillCacheFromJsonString(self, jsonString, maxAge: int, maxAgePictures: Optional[int] = None) -> None:
        self.maxAge = maxAge
        if maxAgePictures is None:
            self.maxAgePictures = maxAge
        else:
            self.maxAgePictures = maxAgePictures

        self.__cache = json.loads(jsonString)
        LOG.info('Reading cache from string')

    def clearCache(self) -> None:
        self.__cache.clear()
        LOG.info('Clearing cache')

    def enableTracker(self) -> None:
        self.__enableTracker = True
        for vehicle in self.vehicles:
            vehicle.enableTracker()

    def disableTracker(self) -> None:
        self.__enableTracker = True
        for vehicle in self.vehicles:
            vehicle.disableTracker()

    def login(self) -> None:
        self.__session.login()

    @property
    def vehicles(self) -> AddressableDict[str, Vehicle]:
        return self.__vehicles

    def update(self, updateCapabilities: bool = True, updatePictures: bool = True, force: bool = False,
               selective: Optional[list[Domain]] = None) -> None:
        self.__elapsed.clear()
        try:
            self.updateVehicles(updateCapabilities=updateCapabilities, updatePictures=updatePictures, force=force, selective=selective)
            self.updateChargingStations(force=force)
        finally:
            self.updateComplete()
            self.__session.cookies.clear()  # Clear cookies to have a fresh session afterwards

    def updateVehicles(self, updateCapabilities: bool = True, updatePictures: bool = True, force: bool = False,  # noqa: C901
                       selective: Optional[list[Domain]] = None) -> None:
        with self.lock:
            catchedRetrievalError = None
            url = 'https://emea.bff.cariad.digital/vehicle/v1/vehicles'
            data = self.fetchData(url, force)
            if data is not None:
                if 'data' in data and data['data']:
                    vins: List[str] = []
                    for vehicleDict in data['data']:
                        if 'vin' not in vehicleDict:
                            break
                        vin: str = vehicleDict['vin']
                        vins.append(vin)
                        try:
                            if vin not in self.__vehicles:
                                vehicle = Vehicle(weConnect=self, vin=vin, parent=self.__vehicles, fromDict=vehicleDict, fixAPI=self.fixAPI,
                                                  updateCapabilities=updateCapabilities, updatePictures=updatePictures, selective=selective,
                                                  enableTracker=self.__enableTracker)
                                self.__vehicles[vin] = vehicle
                            else:
                                self.__vehicles[vin].update(fromDict=vehicleDict, updateCapabilities=updateCapabilities, updatePictures=updatePictures,
                                                            selective=selective)
                        except RetrievalError as retrievalError:
                            catchedRetrievalError = retrievalError
                            LOG.error('Failed to retrieve data for VIN %s: %s', vin, retrievalError)
                    # delete those vins that are not anymore available
                    for vin in [vin for vin in self.__vehicles if vin not in vins]:
                        del self.__vehicles[vin]

                    self.__cache[url] = (data, str(datetime.utcnow()))
            if catchedRetrievalError:
                raise catchedRetrievalError

    def setChargingStationSearchParameters(self, latitude: float, longitude: float, searchRadius: Optional[int] = None, market: Optional[str] = None,
                                           useLocale: Optional[str] = locale.getlocale()[0]) -> None:
        self.latitude = latitude
        self.longitude = longitude
        self.searchRadius = searchRadius
        self.market = market
        self.useLocale = useLocale

    def getChargingStations(self, latitude, longitude, searchRadius=None, market=None, useLocale=None,  # noqa: C901
                            force=False) -> AddressableDict[str, ChargingStation]:
        chargingStationMap: AddressableDict[str, ChargingStation] = AddressableDict(localAddress='', parent=None)
        url: str = f'https://emea.bff.cariad.digital/poi/charging-stations/v2?latitude={latitude}&longitude={longitude}'
        if market is not None:
            url += f'&market={market}'
        if useLocale is not None:
            url += f'&locale={useLocale}'
        if searchRadius is not None:
            url += f'&searchRadius={searchRadius}'
        if self.session.userId is not None:
            url += f'&userId={self.session.userId}'
        data = self.fetchData(url, force)
        if data is not None:
            if 'chargingStations' in data and data['chargingStations']:
                for stationDict in data['chargingStations']:
                    if 'id' not in stationDict:
                        break
                    stationId: str = stationDict['id']
                    station: ChargingStation = ChargingStation(weConnect=self, stationId=stationId, parent=chargingStationMap, fromDict=stationDict,
                                                               fixAPI=self.fixAPI)
                    chargingStationMap[stationId] = station

                self.__cache[url] = (data, str(datetime.utcnow()))
        return chargingStationMap

    def updateChargingStations(self, force: bool = False) -> None:  # noqa: C901 # pylint: disable=too-many-branches
        if self.latitude is not None and self.longitude is not None:
            url: str = f'https://emea.bff.cariad.digital/poi/charging-stations/v2?latitude={self.latitude}&longitude={self.longitude}'
            if self.market is not None:
                url += f'&market={self.market}'
            if self.useLocale is not None:
                url += f'&locale={self.useLocale}'
            if self.searchRadius is not None:
                url += f'&searchRadius={self.searchRadius}'
            if self.session.userId is not None:
                url += f'&userId={self.session.userId}'
            data = self.fetchData(url, force)
            if data is not None:
                if 'chargingStations' in data and data['chargingStations']:
                    ids: List[str] = []
                    for stationDict in data['chargingStations']:
                        if 'id' not in stationDict:
                            break
                        stationId: str = stationDict['id']
                        ids.append(stationId)
                        if stationId not in self.__stations:
                            station: ChargingStation = ChargingStation(weConnect=self, stationId=stationId, parent=self.__stations, fromDict=stationDict,
                                                                       fixAPI=self.fixAPI)
                            self.__stations[stationId] = station
                        else:
                            self.__stations[stationId].update(fromDict=stationDict)
                    # delete those vins that are not anymore available
                    for stationId in [stationId for stationId in ids if stationId not in self.__stations]:
                        del self.__stations[stationId]

                    self.__cache[url] = (data, str(datetime.utcnow()))

    def getLeafChildren(self) -> List[AddressableLeaf]:
        leafChildren = [children for vehicle in self.__vehicles.values() for children in vehicle.getLeafChildren()] \
            + [children for station in self.__stations.values() for children in station.getLeafChildren()]
        if self.__controls.spinControl is not None and self.__controls.spinControl.enabled:
            leafChildren += [self.__controls.spinControl]
        return leafChildren

    def __str__(self) -> str:
        returnString: str = ''
        for vin, vehicle in self.__vehicles.items():
            returnString += f'Vehicle: {vin}\n{vehicle}\n'
        for stationId, station in sorted(self.__stations.items(), key=lambda x: x[1].distance.value, reverse=False):
            returnString += f'Charging Station: {stationId}\n{station}\n'
        return returnString

    def addErrorObserver(self, observer: Callable, errortype: ErrorEventType) -> None:
        self.__errorObservers.add((observer, errortype))
        LOG.debug('%s: Error event observer added for type: %s', self.getGlobalAddress(), errortype)

    def removeErrorObserver(self, observer: Callable, errortype: Optional[ErrorEventType] = None) -> None:
        self.__errorObservers = filter(lambda observerEntry: observerEntry[0] == observer
                                       or (errortype is not None and observerEntry[1] == errortype), self.__errorObservers)

    def getErrorObservers(self, errortype) -> List[Any]:
        return [observerEntry[0] for observerEntry in self.getErrorObserverEntries(errortype)]

    def getErrorObserverEntries(self, errortype: ErrorEventType) -> List[Any]:
        observers: Set[Tuple[Callable, ErrorEventType]] = set()
        for observerEntry in self.__errorObservers:
            observer, observertype = observerEntry
            del observer
            if errortype & observertype:
                observers.add(observerEntry)
        return observers

    def notifyError(self, element, errortype: ErrorEventType, detail: string, message: string = None) -> None:
        observers: List[Callable] = self.getErrorObservers(errortype)
        for observer in observers:
            observer(element=element, errortype=errortype, detail=detail, message=message)
        LOG.debug('%s: Notify called for errors with type: %s for %d observers', self.getGlobalAddress(), errortype, len(observers))

    def recordElapsed(self, elapsed: timedelta) -> None:
        self.__elapsed.append(elapsed)

    def getMinElapsed(self) -> timedelta:
        if len(self.__elapsed) == 0:
            return None
        return min(self.__elapsed)

    def getMaxElapsed(self) -> timedelta:
        if len(self.__elapsed) == 0:
            return None
        return max(self.__elapsed)

    def getAvgElapsed(self) -> timedelta:
        if len(self.__elapsed) == 0:
            return None
        return sum(self.__elapsed, timedelta()) / len(self.__elapsed)

    def getTotalElapsed(self) -> timedelta:
        if len(self.__elapsed) == 0:
            return None
        return sum(self.__elapsed, timedelta())

    def fetchData(self, url, force=False, allowEmpty=False, allowHttpError=False, allowedErrors=None) -> Optional[Dict[str, Any]]:  # noqa: C901
        data: Optional[Dict[str, Any]] = None
        cacheDate: Optional[datetime] = None
        if not force and (self.maxAge is not None and self.cache is not None and url in self.cache):
            data, cacheDateString = self.cache[url]
            cacheDate = datetime.fromisoformat(cacheDateString)
        if data is None or self.maxAge is None \
                or (cacheDate is not None and cacheDate < (datetime.utcnow() - timedelta(seconds=self.maxAge))):
            try:
                statusResponse: requests.Response = self.session.get(url, allow_redirects=False)
                self.recordElapsed(statusResponse.elapsed)
                if statusResponse.status_code in (requests.codes['ok'], requests.codes['multiple_status']):
                    data = statusResponse.json()
                    if self.cache is not None:
                        self.cache[url] = (data, str(datetime.utcnow()))
                elif statusResponse.status_code == requests.codes['too_many_requests']:
                    self.notifyError(self, ErrorEventType.HTTP, str(statusResponse.status_code),
                                     'Could not fetch data due to too many requests from your account')
                    raise TooManyRequestsError('Could not fetch data due to too many requests from your account. '
                                               f'Status Code was: {statusResponse.status_code}')
                elif statusResponse.status_code == requests.codes['unauthorized']:
                    LOG.info('Server asks for new authorization')
                    self.login()
                    statusResponse = self.session.get(url, allow_redirects=False)
                    self.recordElapsed(statusResponse.elapsed)

                    if statusResponse.status_code in (requests.codes['ok'], requests.codes['multiple_status']):
                        data = statusResponse.json()
                        if self.cache is not None:
                            self.cache[url] = (data, str(datetime.utcnow()))
                    elif not allowHttpError or (allowedErrors is not None and statusResponse.status_code not in allowedErrors):
                        self.notifyError(self, ErrorEventType.HTTP, str(statusResponse.status_code), 'Could not fetch data due to server error')
                        raise RetrievalError(f'Could not fetch data even after re-authorization. Status Code was: {statusResponse.status_code}')
                elif not allowHttpError or (allowedErrors is not None and statusResponse.status_code not in allowedErrors):
                    self.notifyError(self, ErrorEventType.HTTP, str(statusResponse.status_code), 'Could not fetch data due to server error')
                    raise RetrievalError(f'Could not fetch data. Status Code was: {statusResponse.status_code}')
            except requests.exceptions.ConnectionError as connectionError:
                self.notifyError(self, ErrorEventType.CONNECTION, 'connection', 'Could not fetch data due to connection problem')
                raise RetrievalError from connectionError
            except requests.exceptions.ChunkedEncodingError as chunkedEncodingError:
                self.notifyError(self, ErrorEventType.CONNECTION, 'chunked encoding error',
                                 'Could not fetch data due to connection problem with chunked encoding')
                raise RetrievalError from chunkedEncodingError
            except requests.exceptions.ReadTimeout as timeoutError:
                self.notifyError(self, ErrorEventType.TIMEOUT, 'timeout', 'Could not fetch data due to timeout')
                raise RetrievalError from timeoutError
            except requests.exceptions.RetryError as retryError:
                raise RetrievalError from retryError
            except requests.exceptions.JSONDecodeError as jsonError:
                if allowEmpty:
                    data = None
                else:
                    self.notifyError(self, ErrorEventType.JSON, 'json', 'Could not fetch data due to error in returned data')
                    raise RetrievalError from jsonError
        return data



from enum import Enum


class Domain(Enum):
    ALL = 'all'
    ALL_CAPABLE = 'allCapable'
    ACCESS = 'access'
    ACTIVEVENTILATION = 'activeventilation'
    AUTOMATION = 'automation'
    AUXILIARY_HEATING = 'auxiliaryheating'
    USER_CAPABILITIES = 'userCapabilities'
    CHARGING = 'charging'
    CHARGING_PROFILES = 'chargingProfiles'
    BATTERY_CHARGING_CARE = 'batteryChargingCare'
    CLIMATISATION = 'climatisation'
    CLIMATISATION_TIMERS = 'climatisationTimers'
    DEPARTURE_TIMERS = 'departureTimers'
    FUEL_STATUS = 'fuelStatus'
    VEHICLE_LIGHTS = 'vehicleLights'
    LV_BATTERY = 'lvBattery'
    READINESS = 'readiness'
    VEHICLE_HEALTH_INSPECTION = 'vehicleHealthInspection'
    VEHICLE_HEALTH_WARNINGS = 'vehicleHealthWarnings'
    OIL_LEVEL = 'oilLevel'
    MEASUREMENTS = 'measurements'
    BATTERY_SUPPORT = 'batterySupport'
    PARKING = 'parking'
    TRIPS = 'trips'

    def __str__(self):
        return self.value



import pytest

from weconnect import addressable


def test_AddressableLeafGetObservers():
    parentAddressableLeaf = addressable.AddressableObject(localAddress='none', parent=None)
    addressableLeaf = addressable.AddressableLeaf(localAddress='none', parent=parentAddressableLeaf)

    def observe1():
        pass

    def observe2():
        pass

    def observe3():
        pass

    def observe4():
        pass

    observerEntries = addressableLeaf.getObserverEntries(addressable.AddressableLeaf.ObserverEvent.ALL)
    assert len(observerEntries) == 0

    addressableLeaf.addObserver(observe1, flag=addressable.AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                priority=addressable.AddressableLeaf.ObserverPriority.INTERNAL_LOW)

    observerEntries = addressableLeaf.getObserverEntries(addressable.AddressableLeaf.ObserverEvent.ALL)
    assert len(observerEntries) == 1

    addressableLeaf.addObserver(observe2, flag=addressable.AddressableLeaf.ObserverEvent.ENABLED,
                                priority=addressable.AddressableLeaf.ObserverPriority.INTERNAL_HIGH)

    observerEntries = addressableLeaf.getObserverEntries(addressable.AddressableLeaf.ObserverEvent.ALL)
    assert len(observerEntries) == 2

    # Add without priority assuming it will have USER_MID as default
    addressableLeaf.addObserver(observe3, flag=addressable.AddressableLeaf.ObserverEvent.ALL)

    observerEntries = addressableLeaf.getObserverEntries(flags=addressable.AddressableLeaf.ObserverEvent.ALL)
    assert len(observerEntries) == 3

    assert observerEntries[0][0] == observe2  # pylint: disable=comparison-with-callable
    assert observerEntries[0][1] == addressable.AddressableLeaf.ObserverEvent.ENABLED
    assert observerEntries[0][2] == addressable.AddressableLeaf.ObserverPriority.INTERNAL_HIGH

    assert observerEntries[1][0] == observe3  # pylint: disable=comparison-with-callable
    assert observerEntries[1][1] == addressable.AddressableLeaf.ObserverEvent.ALL
    assert observerEntries[1][2] == addressable.AddressableLeaf.ObserverPriority.USER_MID

    assert observerEntries[2][0] == observe1  # pylint: disable=comparison-with-callable
    assert observerEntries[2][1] == addressable.AddressableLeaf.ObserverEvent.VALUE_CHANGED
    assert observerEntries[2][2] == addressable.AddressableLeaf.ObserverPriority.INTERNAL_LOW

    observers = addressableLeaf.getObservers(flags=addressable.AddressableLeaf.ObserverEvent.ALL)

    assert len(observers) == 3

    for observerEntry, observer in zip(observerEntries, observers):
        assert observerEntry[0] == observer

    # Now request only a subset of registred observers
    observerEntries = addressableLeaf.getObserverEntries(flags=addressable.AddressableLeaf.ObserverEvent.VALUE_CHANGED)
    assert len(observerEntries) == 2

    # Observer for post update only
    observerEntries = addressableLeaf.getObserverEntries(addressable.AddressableLeaf.ObserverEvent.ALL, onUpdateComplete=True)
    assert len(observerEntries) == 0

    addressableLeaf.addObserver(observe4, flag=addressable.AddressableLeaf.ObserverEvent.ENABLED,
                                priority=addressable.AddressableLeaf.ObserverPriority.INTERNAL_HIGH, onUpdateComplete=True)

    observerEntries = addressableLeaf.getObserverEntries(addressable.AddressableLeaf.ObserverEvent.ALL)
    assert len(observerEntries) == 3
    observerEntries = addressableLeaf.getObserverEntries(addressable.AddressableLeaf.ObserverEvent.ALL, onUpdateComplete=True)
    assert len(observerEntries) == 1


def test_AddressableLeafParents():
    parentAddressableLeaf = addressable.AddressableObject(localAddress='none', parent=None)
    addressableLeaf = addressable.AddressableLeaf(localAddress='none', parent=parentAddressableLeaf)

    assert parentAddressableLeaf.enabled is False
    assert addressableLeaf.enabled is False

    leafChildren = parentAddressableLeaf.getLeafChildren()
    assert len(leafChildren) == 0

    addressableLeaf.enabled = True

    assert parentAddressableLeaf.enabled is True
    assert addressableLeaf.enabled is True

    addressableLeaf.enabled = True
    assert addressableLeaf.enabled is True

    recursiveChildren = parentAddressableLeaf.getRecursiveChildren()
    assert len(recursiveChildren) == 2  # This is two as it is the oarent as well as the child
    leafChildren = parentAddressableLeaf.getLeafChildren()
    assert len(leafChildren) == 1

    addressableLeaf.enabled = False
    assert addressableLeaf.enabled is False

    del addressableLeaf
    del parentAddressableLeaf


def test_AddressableLeafAdresses():
    parentAddressableLeaf = addressable.AddressableObject(localAddress='parent', parent=None)
    addressableLeaf = addressable.AddressableLeaf(localAddress='child', parent=parentAddressableLeaf)

    getterAddress = addressableLeaf.localAddress
    localAdress = addressableLeaf.getLocalAddress()
    globalAdress = addressableLeaf.getGlobalAddress()

    assert getterAddress == localAdress
    assert localAdress == 'child'
    assert globalAdress == 'parent/child'

    addressableLeaf.localAddress = 'newChild'

    localAdress = addressableLeaf.getLocalAddress()
    globalAdress = addressableLeaf.getGlobalAddress()
    assert localAdress == 'newChild'
    assert globalAdress == 'parent/newChild'


def test_AddressableLeafParent():
    parentAddressableLeaf = addressable.AddressableObject(localAddress='parent', parent=None)
    addressableLeaf = addressable.AddressableLeaf(localAddress='child', parent=parentAddressableLeaf)

    assert addressableLeaf.parent == parentAddressableLeaf

    assert addressableLeaf.getRoot() == parentAddressableLeaf
    assert parentAddressableLeaf.getRoot() == parentAddressableLeaf

    newParentAddressableLeaf = addressable.AddressableObject(localAddress='newparent', parent=None)
    addressableLeaf.parent = newParentAddressableLeaf

    assert addressableLeaf.parent == newParentAddressableLeaf


#  TODO: By string not complete implementation yet
def test_AddressableByAddressString():
    root = addressable.AddressableObject(localAddress='root', parent=None)
    parent = addressable.AddressableObject(localAddress='parent', parent=root)
    leaf = addressable.AddressableLeaf(localAddress='leaf', parent=parent)

    assert leaf.getByAddressString('anything') is False
    assert leaf.getByAddressString('leaf') == leaf
    assert leaf.getByAddressString('..') == parent
    assert leaf.getByAddressString('/') == root

    assert parent.getByAddressString('anything') is False
    assert parent.getByAddressString('parent') == parent
    assert parent.getByAddressString('..') == root
    assert parent.getByAddressString('/') == root

    assert root.getByAddressString('anything') is False
    assert root.getByAddressString('root') == root
    assert root.getByAddressString('..') == root
    assert root.getByAddressString('/') == root


def test_AdressableAttribute():
    attribute = addressable.AddressableAttribute(localAddress='test', parent=None, value=None, valueType=str, lastUpdateFromCar=None)
    assert attribute.value is None

    attribute = addressable.AddressableAttribute(localAddress='test', parent=None, value='value', valueType=str, lastUpdateFromCar=None)
    assert attribute.value == 'value'

    with pytest.raises(NotImplementedError):
        attribute.value = 'newValue'


def test_AdressableObjectLeafChildren():
    addressableObject = addressable.AddressableObject(localAddress='parent', parent=None)
    childAddressableLeaf1 = addressable.AddressableAttribute(localAddress='child1', parent=addressableObject, value=None, valueType=str, lastUpdateFromCar=None)
    childAddressableLeaf2 = addressable.AddressableAttribute(localAddress='child2', parent=addressableObject, value=None, valueType=str, lastUpdateFromCar=None)

    children = addressableObject.getLeafChildren()
    assert len(children) == 0

    addressableObject.enabled = True
    children = addressableObject.getLeafChildren()
    assert len(children) == 1
    assert children[0] == addressableObject

    childAddressableLeaf1.setValueWithCarTime(newValue='test1')

    children = addressableObject.getLeafChildren()
    assert len(children) == 1
    children = addressableObject.getRecursiveChildren()
    assert len(children) == 2
    assert children[0] == addressableObject
    assert children[1] == childAddressableLeaf1

    childAddressableLeaf2.setValueWithCarTime(newValue='test2')

    children = addressableObject.getLeafChildren()
    assert len(children) == 2
    children = addressableObject.getRecursiveChildren()
    assert len(children) == 3
    assert children[0] == addressableObject
    assert children[1] == childAddressableLeaf1
    assert children[2] == childAddressableLeaf2



import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class MaintenanceStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.inspectionDue_days = AddressableAttribute(localAddress='inspectionDue_days', parent=self, value=None, valueType=int)
        self.inspectionDue_km = AddressableAttribute(localAddress='inspectionDue_km', value=None, parent=self, valueType=int)
        self.mileage_km = AddressableAttribute(localAddress='mileage_km', value=None, parent=self, valueType=int)
        self.oilServiceDue_days = AddressableAttribute(localAddress='oilServiceDue_days', value=None, parent=self, valueType=int)
        self.oilServiceDue_km = AddressableAttribute(localAddress='oilServiceDue_km', value=None, parent=self, valueType=int)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update maintenance status from dict')

        if 'value' in fromDict:
            self.inspectionDue_days.fromDict(fromDict['value'], 'inspectionDue_days')
            self.inspectionDue_km.fromDict(fromDict['value'], 'inspectionDue_km')

            if 'mileage_km' in fromDict['value']:
                mileage_km = int(fromDict['value']['mileage_km'])
                if self.fixAPI and mileage_km == 0x7FFFFFFF:
                    mileage_km = None
                    LOG.info('%s: Attribute mileage_km was error value 0x7FFFFFFF. Setting error state instead'
                             ' of 2147483647 km.', self.getGlobalAddress())
                self.mileage_km.setValueWithCarTime(mileage_km, lastUpdateFromCar=None, fromServer=True)
            else:
                self.mileage_km.enabled = False
            self.oilServiceDue_days.fromDict(fromDict['value'], 'oilServiceDue_days')
            self.oilServiceDue_km.fromDict(fromDict['value'], 'oilServiceDue_km')
        else:
            self.inspectionDue_days.enabled = False
            self.inspectionDue_km.enabled = False
            self.mileage_km.enabled = False
            self.oilServiceDue_days.enabled = False
            self.oilServiceDue_km.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + [
                                                                'inspectionDue_days',
                                                                'inspectionDue_km',
                                                                'mileage_km',
                                                                'oilServiceDue_days',
                                                                'oilServiceDue_km'
                                                            ]))

    def __str__(self):
        string = super().__str__()
        if self.mileage_km.enabled:
            string += f'\n\tCurrent milage: {self.mileage_km.value} km'
        if self.inspectionDue_days.enabled:
            string += f'\n\tInspection due in: {self.inspectionDue_days.value} days'
        if self.inspectionDue_km.enabled:
            string += f'\n\tInspection due in: {self.inspectionDue_km.value} km'
        if self.oilServiceDue_km.enabled:
            string += f'\n\tOil service in: {self.oilServiceDue_km.value} km'
        if self.oilServiceDue_days.enabled:
            string += f'\n\tOil service in: {self.oilServiceDue_days.value} days'
        return string



import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus
from weconnect.elements.enums import EngineType, CarType

LOG = logging.getLogger("weconnect")


class FuelLevelStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.currentFuelLevel_pct = AddressableAttribute(localAddress='currentFuelLevel_pct', parent=self, value=None, valueType=int)
        self.currentSOC_pct = AddressableAttribute(localAddress='currentSOC_pct', parent=self, value=None, valueType=int)
        self.primaryEngineType = AddressableAttribute(localAddress='primaryEngineType', parent=self, value=None, valueType=EngineType)
        self.secondaryEngineType = AddressableAttribute(localAddress='secondaryEngineType', parent=self, value=None, valueType=EngineType)
        self.carType = AddressableAttribute(localAddress='carType', parent=self, value=None, valueType=CarType)

        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update fuel level status from dict')

        if 'value' in fromDict:
            self.currentFuelLevel_pct.fromDict(fromDict['value'], 'currentFuelLevel_pct')
            self.currentSOC_pct.fromDict(fromDict['value'], 'currentSOC_pct')
            self.primaryEngineType.fromDict(fromDict['value'], 'primaryEngineType')
            self.secondaryEngineType.fromDict(fromDict['value'], 'secondaryEngineType')
            self.carType.fromDict(fromDict['value'], 'carType')
        else:
            self.currentFuelLevel_pct.enabled = False
            self.currentSOC_pct.enabled = False
            self.primaryEngineType.enabled = False
            self.secondaryEngineType.enabled = False
            self.carType.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + ['currentFuelLevel_pct',
                                                               'currentSOC_pct',
                                                               'primaryEngineType',
                                                               'secondaryEngineType',
                                                               'carType']))

    def __str__(self):
        string = super().__str__()
        if self.carType.enabled:
            string += f'\n\tCar Type: {self.carType.value.value}'  # pylint: disable=no-member
        if self.primaryEngineType.enabled:
            string += f'\n\tPrimary Engine: {self.primaryEngineType.value.value}'
        if self.secondaryEngineType.enabled:
            string += f'\n\tSecondary Engine: {self.secondaryEngineType.value.value}'
        if self.currentFuelLevel_pct.enabled:
            string += f'\n\tCurrent Fuel Level: {self.currentFuelLevel_pct.value}%'
        if self.currentSOC_pct.enabled:
            string += f'\n\tCurrent Charge Level: {self.currentSOC_pct.value}%'
        return string



import json
import logging
import requests

from urllib.parse import parse_qsl, urlparse

from oauthlib.common import add_params_to_uri, generate_nonce, to_unicode
from oauthlib.oauth2 import InsecureTransportError
from oauthlib.oauth2 import is_secure_transport

from requests.models import CaseInsensitiveDict
from weconnect.auth.openid_session import AccessType

from weconnect.auth.vw_web_session import VWWebSession
from weconnect.errors import AuthentificationError, RetrievalError, TemporaryAuthentificationError


LOG = logging.getLogger("weconnect")


class WeConnectSession(VWWebSession):
    def __init__(self, sessionuser, **kwargs):
        super(WeConnectSession, self).__init__(client_id='a24fba63-34b3-4d43-b181-942111e6bda8@apps_vw-dilab_com',
                                               refresh_url='https://identity.vwgroup.io/oidc/v1/token',
                                               scope='openid profile badge cars dealers vin',
                                               redirect_uri='weconnect://authenticated',
                                               state=None,
                                               sessionuser=sessionuser,
                                               **kwargs)

        self.headers = CaseInsensitiveDict({
            'accept': '*/*',
            'content-type': 'application/json',
            'content-version': '1',
            'x-newrelic-id': 'VgAEWV9QDRAEXFlRAAYPUA==',
            'user-agent': 'Volkswagen/3.51.1-android/14',
            'accept-language': 'de-de',
            'Cache-Control': 'no-cache',
            'Pragma': 'no-cache',
            'x-android-package-name': 'com.volkswagen.weconnect'
        })

    def request(
        self,
        method,
        url,
        data=None,
        headers=None,
        withhold_token=False,
        access_type=AccessType.ACCESS,
        token=None,
        timeout=None,
        **kwargs
    ):
        """Intercept all requests and add weconnect-trace-id header."""

        import secrets
        traceId = secrets.token_hex(16)
        weConnectTraceId = (traceId[:8] + '-' + traceId[8:12] + '-' + traceId[12:16] + '-' + traceId[16:20] + '-' + traceId[20:]).upper()
        headers = headers or {}
        headers['weconnect-trace-id'] = weConnectTraceId

        return super(WeConnectSession, self).request(
            method, url, headers=headers, data=data, withhold_token=withhold_token, access_type=access_type, token=token, timeout=timeout, **kwargs
        )

    def login(self):
        super(WeConnectSession, self).login()
        authorizationUrl = self.authorizationUrl(url='https://identity.vwgroup.io/oidc/v1/authorize')
        response = self.doWebAuth(authorizationUrl)
        self.fetchTokens('https://emea.bff.cariad.digital/user-login/login/v1',
                         authorization_response=response
                         )

    def refresh(self):
        self.refreshTokens(
            'https://emea.bff.cariad.digital/login/v1/idk/token',
        )

    def authorizationUrl(self, url, state=None, **kwargs):
        if state is not None:
            raise AuthentificationError('Do not provide state')

        params = [(('redirect_uri', self.redirect_uri)),
                  (('nonce', generate_nonce()))]

        authUrl = add_params_to_uri('https://emea.bff.cariad.digital/user-login/v1/authorize', params)

        tryLoginResponse: requests.Response = self.get(authUrl, allow_redirects=False, access_type=AccessType.NONE)
        redirect = tryLoginResponse.headers['Location']
        query = urlparse(redirect).query
        params = dict(parse_qsl(query))
        if 'state' in params:
            self.state = params.get('state')

        return redirect

    def clearTokens(self) -> None:
        """
        Clear all stored tokens to force a fresh login.
        
        This method is useful when the server requests new authorization
        and we need to clear invalid/expired tokens.
        """
        LOG.info("Clearing all stored tokens")
        self.token = None
        LOG.debug("All tokens cleared successfully")

    def fetchTokens(
        self,
        token_url,
        authorization_response=None,
        **kwargs
    ):
        self.parseFromFragment(authorization_response)

        if all(key in self.token for key in ('state', 'id_token', 'access_token', 'code')):
            body: str = json.dumps(
                {
                    'state': self.token['state'],
                    'id_token': self.token['id_token'],
                    'redirect_uri': self.redirect_uri,
                    'region': 'emea',
                    'access_token': self.token['access_token'],
                    'authorizationCode': self.token['code'],
                })

            loginHeadersForm: CaseInsensitiveDict = self.headers
            loginHeadersForm['accept'] = 'application/json'

            tokenResponse = self.post(token_url, headers=loginHeadersForm, data=body, allow_redirects=False, access_type=AccessType.ID)
            if tokenResponse.status_code != requests.codes['ok']:
                raise TemporaryAuthentificationError(f'Token could not be fetched due to temporary WeConnect failure: {tokenResponse.status_code}')
            token = self.parseFromBody(tokenResponse.text)

            # Ensure the token is properly stored in the session
            if token is not None:
                self.token = token  # Explicitly store the token
                LOG.debug(f"Successfully fetched tokens. Access token expires in: {token.get('expires_in', 'unknown')} seconds")
                LOG.debug(f"Refresh token available: {'refresh_token' in token}")
                # Verify critical tokens are present
                if not all(key in token for key in ('access_token', 'id_token', 'refresh_token')):
                    LOG.warning("Some expected tokens are missing from the response")
            else:
                LOG.error("Token parsing returned None")

            return token
        else:
            LOG.error("Authorization response missing required tokens")
            return None

    def parseFromBody(self, token_response, state=None):
        try:
            token = json.loads(token_response)
        except json.decoder.JSONDecodeError:
            raise TemporaryAuthentificationError('Token could not be refreshed due to temporary WeConnect failure: json could not be decoded')
        if 'accessToken' in token:
            token['access_token'] = token.pop('accessToken')
        if 'idToken' in token:
            token['id_token'] = token.pop('idToken')
        if 'refreshToken' in token:
            token['refresh_token'] = token.pop('refreshToken')
        fixedTokenresponse = to_unicode(json.dumps(token)).encode("utf-8")
        parsedToken = super(WeConnectSession, self).parseFromBody(token_response=fixedTokenresponse, state=state)
        # Ensure the token is stored in the session object
        self.token = parsedToken
        return parsedToken

    def refreshTokens(
        self,
        token_url,
        refresh_token=None,
        auth=None,
        timeout=None,
        headers=None,
        verify=True,
        proxies=None,
        **kwargs
    ):
        LOG.info('Refreshing tokens')
        if not token_url:
            raise ValueError("No token endpoint set for auto_refresh.")

        if not is_secure_transport(token_url):
            raise InsecureTransportError()

        refresh_token = refresh_token or self.refreshToken

        if headers is None:
            headers = self.headers

        # First try to get from the current token property, then fall back to stored token
        if refresh_token is None:
            refresh_token = self.refresh_token
            # If still None, try to get from the token dict directly
            if refresh_token is None and self.token is not None:
                refresh_token = self.token.get('refresh_token')

        if not refresh_token:
            raise AuthenticationError('No refresh token available. Please log in again.')

        # Create headers matching the examples format
        tHeaders = {
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": "Volkswagen/3.51.1-android/14",
            "x-android-package-name": "com.volkswagen.weconnect",
        }

        # Create form data body matching the examples format
        body = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self.client_id,
        }

        # Request new tokens using POST with form data
        tokenResponse = self.post(
            token_url,
            data=body,
            headers=tHeaders,
            timeout=timeout,
            verify=verify,
            proxies=proxies,
        )
        if tokenResponse.status_code == requests.codes['unauthorized']:
            LOG.error('Token refresh failed with 401 - server requests new authorization. Refresh token may be expired or invalid.')
            raise AuthentificationError('Refreshing tokens failed: Server requests new authorization. Please log in again.')
        elif tokenResponse.status_code in (requests.codes['internal_server_error'], requests.codes['service_unavailable'], requests.codes['gateway_timeout']):
            raise TemporaryAuthentificationError(f'Token could not be refreshed due to temporary WeConnect failure: {tokenResponse.status_code}')
        elif tokenResponse.status_code == requests.codes['ok']:
            newToken = self.parseFromBody(tokenResponse.text)
            if newToken is not None and "refresh_token" not in newToken:
                LOG.debug("No new refresh token given. Re-using old.")
                self.token["refresh_token"] = refresh_token
                # Update the token property as well
                self.token = newToken
            return newToken
        else:
            raise RetrievalError(f'Status Code from WeConnect while refreshing tokens was: {tokenResponse.status_code}')
